#ifndef BASICOBJECTS_H
#define BASICOBJECTS_H

#include "nodeptr.h"
#include "BasicTypes.h"


/**
 * Dieses Modul definiert eine Schnittstelle zur Erstellung und Aktualisierung 
 * einer visuellen Darstellung. Konzeptionell werden dazu Objekte auf ein Canvas gezeichnet.
 * Jedes Objekt hat einen eindeutigen Namen, gebildet aus dem
 * uebergebenen context-Pointer und seinem "partname". Ferner
 * bekommt jedes Objekt nach seiner Erzeugung eine Canvas-ID.
 *
 * Ueber den Namen werden Objekte, die schon existieren identifiziert
 * und bei einem erneuten AG-Durchlauf nur noch geaendert. Objekte, 
 * die nach einem Durchlauf nicht mehr existieren, werden geloescht.
 * 
 * Die Tiefe eines Objekts bestimmt seine Sichtbarkeit, wenn mehrere
 * Objekte sich ueberdecken: Objekte mit hoeherem Wert liegen weiter 
 * zum Betrachter und ueberdecken Objekte mit niedrigerem Wert.
 * 
 * Dieses Modul enthaelt zusaetzlich einige Hilfsfunktionen, z.B. Erzeugung
 * von "partnamen" und das Abfragen von Knotentypen des Baumes.
 */



extern char VLStringBuffer[2000];
#define PTRVAL(x) (*x)
#define GETVLSTRING(printf) (printf, VLString(VLStringBuffer))
#define VLDYNSTRING(charptr) (VLString(charptr).getCharPtr())

/* Wenn Baumknoten mehrere getrennte Darstellungen in einer Sicht
   besitzen (realisiert durch spezielle Repraesentanten-Baumknoten),
   dann muessen die visuellen Objekte anhand des Repraesentantenknotens
   unterschieden werden */

#define PARTNAME1(nodeptr,str) GETVLSTRING(sprintf(VLStringBuffer,"%p_%s",nodeptr,str))
#define PARTNAME2(str) PARTNAME1(_currn,str)

/**RAW-PUBLIC*/
#define INSPOINT_ABOVE   1
#define INSPOINT_BELOW   2 
#define INSPOINT_SON     3
#define INSPOINT_REPLACE 4
#define INSPOINT_END     5 

#define INSPOINT_NORMAL  'b'
#define INSPOINT_INSONLY 'i'
#define INSPOINT_MOVONLY 'm'
/**END-PUBLIC*/

/**PUBLIC*/
void VLSetWindowTitle(VLCharPtr title);

/* Setzt den Titel des Sicht-Fensters. */


/**PUBLIC*/
VLString VLSymName(NODEPTR p);

/* Liefert zum gegebenen Baumknoten den Namen des Symbols auf der Linken Seite der
 * entsprechenden Produktion.
 */


/**PUBLIC*/
int VLObjectCreate(NODEPTR context, VLCharPtr partname, VLCharPtr type, int depth, 
        VLPointArray *coords, VLListPtr attributes);
int VLObjectCreate(NODEPTR context, VLCharPtr partname, VLCharPtr type, int depth, 
        VLPointArray *coords, VLCharPtr attributes = (VLCharPtr)0);
int VLObjectCreate(NODEPTR context, VLCharPtr partname, VLCharPtr type, int depth, 
        VLPointArray *coords, Tcl_Obj *attributes);
/* Erzeugt ein neues Objekt fuer den Knoten [context]. [partname]
 * ist ein fuer [context] eindeutiger, das Objekt
 * identifizierender Name.  [type] ist der name eines
 * Canvastyps. [depth] die Tiefe des neuen Objekts.  [coords] die
 * Koordinaten des Objekts, je nach Canvastyp.  [attributes] ist ein
 * String, der Attribute des neuen Objekts einstellt.
 *
 * [type] ist der Typ eines Canvasobjekts. Gueltige Canvasobjekte sind:
 * arc, bitmap, image, line, oval, polygon, rectangle, text, (window)
 * Weitere Typen koennen hinzugefuegt werden: siehe manpage 'Tk_CanvasCreateItemType'.
 * Die Anzahl der Punkte, die fuer eine Koordinate gebraucht werden, ist
 * durch den Typ bestimmt: Siehe manpage 'canvas'
 * Attributes ist eine Liste von Attributen fuer den Canvastyp, siehe
 * ebenfalls die manpage 'canvas'.
 */


/**PUBLIC*/
VLCharPtr VLFormatPartname(VLCharPtr partname, int id);
/* Formatiert einen Partname bestehend aus einem Praefix und einer
 * Zahl. Es wird hierbei ein statischer String mit einer maximalen
 * Laenge verwendet, also Vorsicht! */


/**PUBLIC*/
int VLFixInspointCreate(NODEPTR context, VLCharPtr kind, VLPointArray *coords, 
        int where, VLCharPtr partname = (VLCharPtr)0, char action = INSPOINT_NORMAL, 
        VLListPtr posattr = (VLListPtr)0);

/* Erzeugt eine Einfuegestelle ohne Positionierungsfreiheiten.
 * 
 * !param context: Einfuegekontext
 * !param kind: Ist ein Name fuer
 *              Einfgestellenarten. Neue Einfuegestellenarten
 *              koennen durch Aufruf einer Tcl-Prozedur
 *              (::visual::create_inspoint_kind) erzeugt werden.  
 * !param where: Bestimmt die genaue Einfuegestelle.
 *               Es sind die [INSPOINT_*]-Konstanten oben einzusetzen.
 * !param partname: kann verwendet werden, um zu einem Kontext im Strukturbaum 
 *               mehr als eine Einfuegestelle zu erzeugen.
 * 
 * !returns: Die ID der Einfuegestelle
 *
 * !seealso: VLMovableInspointCreate
 */


/**PUBLIC*/
int VLMovableInspointCreate(NODEPTR context, VLCharPtr kind, VLPointArray *coords, 
        int where, VLCharPtr partname = (VLCharPtr)0, 
        VLPointArrayPtr bbox = (VLPointArrayPtr)0, int depth = 0, char action = INSPOINT_NORMAL, VLCharPtr posattr = (VLCharPtr)0);


/* Erzeugt eine Einfuegestelle, bei der waehrend des Einfuegens gleichzeitig eine
 * Platzierungsentscheidung getroffen wird.
 *
 * !param bbox: Position der Einfuegestelle (default: Das gesamte Fenster) 
 *
 * !param posattr: Eine Liste bestehend aus Name des Attributs und optional 
 *                 den X- und Y-Wert der von der Cursorposition vor Setzen des Attributs
 *                 subtrahiert werden soll, um die Position zu einer Relativposition zu
 *                 machen. 
 * 
 * !returns: Die ID der Einfuegestelle
 *
 * !seealso: VLFixInspointCreate
 * !seealso: VLRepositionCreate
 */


/**PUBLIC*/
int VLRepositionCreate(NODEPTR context, VLCharPtr kind, VLPointArray *coords, 
        VLListPtr symbolnames, VLCharPtr partname = (VLCharPtr)0, 
        VLPointArrayPtr bbox = (VLPointArrayPtr)0, int depth = 0, VLCharPtr posattr = (VLCharPtr)0,
        VLPointArrayPtr orignin = (VLPointArrayPtr)0);

/* Ermoeglicht, die Position eines gafischen Objektes zu veraendern, ohne eine
 * dazu eine Einfuegestelle benutzen zu muessen.
 *
 * !seealso: VLMovableInspointCreate
 */


/**PUBLIC*/
void VLSetInspointProc(int inspointid, VLCharPtr insertcmd, VLListPtr args = (VLListPtr) 0);
/* Normalerweise wird zum Einfuegen neuer Knoten in einer
 * Einfuegestelle die Prozedur [::vl::insert] benutzt. In einigen
 * Situationen kann es aber wuenschenswert sein, das
 * Default-Verhalten zu aendern, beispielsweise wenn
 * zusaetzliche Attribute gesetzt werden sollen. Hierzu kann mit
 * dieser Prozedur das Default-Verhalten geaendert werden.  Die
 * Funktion [insertcmd] wird mit folgenden Parametern aufgerufen:
 * [insertcmd fathernode index newnode replace insertargs].
 * [fathernode], [index], [newnode] und [replace] sind die Originalparameter, mit denen
 * [::vl::insert] aufgerufen wuerde. [insertargs] sind die weiteren
 * Argumente, die bei dieser Prozedur angegeben wurden.  
 */


void VLSetPosAttrProc(int inspointid, VLCharPtr posattrproc, VLListPtr args = (VLListPtr) 0);

/**PUBLIC*/
void VLConstrainInspoint(int inspointid, NODEPTR root, int tree = 1);

/* Einfuegestelle ist nur anwaehlbar fuer Knoten, die ein Sohn von root  
 * sind. Falls tree 0 ist, so muss root ein Listenkontext sein. Nur Elemente
 * dieser Liste lassen sich dann noch zum angegebenen Einfuegepunkt verschieben.
 */

/**PUBLIC*/
void VLAddEditContext(VLNodePtr visualContext, VLNodePtr editContext);

/* Mit dieser Funktion koennen neue Editier-Contexte zu einem visuellen Objekt hinzugefuegt werden,
 * Diese erscheinen dann im Contextmenue. Dadurch lassen sich also persistente Attrbute eines 
 * anderen Baumknotens editieren. Dies ist sinnvoll, wenn es sich bei dem Objekt
 * um eine Sicht oder um eine Instanz eines anderen Objektes handelt. 
 */


/**PUBLIC*/
void VLContextedit(NODEPTR context, VLCharPtr label, VLCharPtr attribute, VLValuePtr value);
void VLContextedit(NODEPTR context, VLCharPtr label, VLListPtr command);

/* Falls ein Teil der Darstellung von [context] mit der rechten Maustaste 
 * angewaehlt wird, so wird ein Contextmenue gezeigt. Dieses enthaelt als
 * Default-Operationen den Aufruf der Editoren und Delete. Weiterhin koennen
 * Attribut-Setz-Operationen und Befehle eingetragen werden. Sie werden mit [label]
 * im Menue angezeigt. Wenn dieser Menuepunkt angewaehlt wird, wird das Attribut [attribute]
 * von [context] auf den Wert [value] gesetzt. 
 */


/**PUBLIC*/
int VLDragCreateEx(NODEPTR context, VLCharPtr partname, VLCharPtr type, 
                   VLPointArray *coords, VLCharPtr kind, VLCharPtr attrname, 
                   VLCharPtr associatecmd = (VLCharPtr) 0, VLList *args = (VLList *)0);

inline int VLDragCreate(NODEPTR context, VLCharPtr partname, VLPointArray *coords, 
                        VLCharPtr kind, VLCharPtr attrname, 
                        VLCharPtr associatecmd = (VLCharPtr) 0, 
                        VLList *args = (VLList *)0)
    {
        return VLDragCreateEx(context, partname, "rectangle", coords, kind, attrname, 
                              associatecmd, args);
    }
 
/* Erzeugt ein neues, unsichtbares Rechteck, Kreis oder Polygon,
 * das die Herstellung einer Beziehung zu einem Punkt erlaubt.
 *
 * !param context:  Definiert den Drag-Kontext
 * !param partname: Bildet zusammen mit [context] eine eindeutige Bezeichnung fuer das
 *                  Rechteck (genau wie bei [VLCreateObjekt])
 * !param coords:   Die Koordinaten des Rechtecks. Werden nur 2
 *                  Koordinaten angegeben, so sind diese das Zentrum eines 4x4-Pixel
 *                  grossen Rechtecks
 * !param kind:     Die durch VLDragCreate erzeugen Beziehungen
 *                  haben den Namen 'kind' und koennen nur zu den Objekten
 *                  hergestellt werden, die den gleichen Namen haben
 * !param attrname: Wird eine Beziehung hergestellt, so wird das Attribut [attrname] von
 *                  [context] durch eine neue Beziehung ersetzt. [associatecmd] und [args]
 *                  koennen angegeben werden, um das Default-Verhalten des
 *                  Herstellens einer Beziehung durch ein neues Verhalten zu ersetzen
 * !param associatecmd: Wenn angegeben, wird mit folgenden Parametern
 *                  aufgerufen: associatecmd context attrname dropcontext args dropargs
 * !param args:     Benutzerdefinierte Parameter fuer [associatecmd].
 *
 * !seealso: VLSetDrop
 */


/**PUBLIC*/
void VLSetDrop(int objectid, VLCharPtr kind, VLPropertyPtr property, VLList *args = (VLList *)0);
/* Veraendert ein erzeugtes Objekt so, dass es eine
 * Zielstelle fuer die Spezifikation einer Beziehung ist.
 *
 * !param objectid: Eine durch [VLObjectCreate] zurueckgegebene id.
 * !param kind: passt zu dem [kind]-Parameter von [VLDragCreateEx]
 * !param property: markiert die bei DragCreate zu setzende Beziehung
 * !param args: kann weitere Argumente fuer die Prozedur
 *              [associatecmd] von [DragCreate] enthalten.
 *
 * !seealso: VLDragCreateEx
 * !seealso: VLObjectCreate
 */

/**PUBLIC*/
VLPointArray *VLMeasureText(char *text, char *fontname = (char *)0);
/* Gibt ein PointArray zurueck, dass im 1. Punkt die Breite und Hoehe
 * enthaelt, die 'text' haben wuerde, wenn er mit dem angegebenen Font
 * erzeugt werden wuerde. */


/**PUBLIC*/
VLPointArray *VLMeasureImage(char *name);
/* Liefert die Groesse eines Bildes */

/**PUBLIC*/
VLString VLGetImageId(char *);
/* Liefert die ID eines Bildes. Wofuer war die denn eigentlich gut??? */


/**RAW-PUBLIC*/
#define RESIZER_TOPLEFT         0
#define RESIZER_TOP             1
#define RESIZER_TOPRIGHT        2
#define RESIZER_RIGHT           3
#define RESIZER_BOTTOMRIGHT     4
#define RESIZER_BOTTOM          5
#define RESIZER_BOTTOMLEFT      6
#define RESIZER_LEFT            7
/**END-PUBLIC*/


/**PUBLIC*/     
void VLSetResizer(int objectid, int corner, VLCharPtr attrname);
/* Macht ein normales (mit [VLCreateObjekt] erzeugtes) visuelles Objekt zu einem
 * Resize-Knopf. Hierzu wird es unsichtbar gemacht (mit [-fill {} -outline {}])
 * und bei passender Gelegenheit (selektion des Kontextes) aktiviert 
 * (auch durch Setzen von [-fill] und [-outline]). Aus diesem Grund koennen nur die
 * Canvasobjekte fuer Resizer verwendet werden, die [-fill] und [-outline] verstehen.
 * Beispielsweise [rectangle], [oval] und [polygon].

 * !param objectid: Die ID des mit [VLCreateObjekt] erzeugen Objekts.
 * !param corner:   Definiert, an welcher Stelle der Resizer-Knopf
 *                  positioniert werden soll. Siehe auch die
 *                  [RESIZER_*] - Konstanten
 * !param attrname: Name des Attributs, das die Groesse empfaengt.
 *                  Es wird der Kontext zum setzen verwendet, der 
 *                  beim Objekt [objectid] angegeben wurde.
 */

/**PUBLIC*/     
void VLCreateHandle(int objectid, VLCharPtr procname, VLListPtr param);
/* ???. Keine Ahnung, wofuer das hier gut ist. */

/**PUBLIC*/     
void VLSetEditable(int objectid, VLCharPtr attrname);
/* Macht einen Text editierbar. Editiert ein Benutzer den Text, so wird bei dem
 * Knoten, zu dem es gehoert, das Attribut "attrname" auf den eingegebenen Text
 * gesetzt. */

/**PUBLIC*/
VLString VLFile(VLCharPtr filename);
/* Haengt den Pfadnamen des Library-Verzeichnisses vorne an den Namen an. */

#endif
