#include "BasicTypes.h"
#include "tcl_vpt.h"
#include "err.h"

#define assert(ex) {if (!(ex)){ \
  (void)fprintf(stderr, "Assertion failed: file %s, line %d\n", \
                __FILE__, __LINE__); exit(2);}}


VLAutoDelete *VLAutoDelete::last_allocated = (VLAutoDelete *)0;
VLAutoDelete *VLAutoDelete::all_allocated = (VLAutoDelete *)0;


VLAutoDelete::VLAutoDelete():  m_next(0)
{
}

VLAutoDelete::~VLAutoDelete()
{
}


void *VLAutoDelete::operator new(size_t len)
{
    void *mem = Tcl_Alloc(len);

    if (last_allocated)
    {
        last_allocated->m_next = all_allocated;
        all_allocated = last_allocated;
    }
    last_allocated = (VLAutoDelete *)mem;

    return mem;
}

void VLAutoDelete::operator delete(void *mem)
{
    fprintf (stderr, "*** VLAutoDelete::operator delete called\n");
}

void VLAutoDelete::ReleaseMemory()
{
    VLAutoDelete *current;

    if (last_allocated)
    {
        current = last_allocated;
        last_allocated = (VLAutoDelete *)0;
        Tcl_Free((char *)current);
    }

    while (all_allocated)
    {
        current = all_allocated;
        all_allocated = all_allocated->m_next;

        current->VLAutoDelete::~VLAutoDelete();     // Explicit Destructor Call
        Tcl_Free((char *)current);
    }
}




Tcl_Obj *VLValue::eval(void)
{
        int ret = Tcl_GlobalEvalObj(TclInterp, m_val);
        if (ret != TCL_OK) 
        {
                ReportTclError();
                return 0;
        }
        return Tcl_GetObjResult(TclInterp);
}
        


VLValue::VLValue():  m_val(0)
{
}

VLValue::VLValue(const VLValue &v): 
    m_val(v.m_val)
{
    if (m_val)
        Tcl_IncrRefCount(m_val);
}

VLValue::VLValue(Tcl_Obj *val): m_val(val)
{
    if (m_val)
        Tcl_IncrRefCount(m_val);
}

void VLValue::setResult(void)
{
  Tcl_SetObjResult(TclInterp, m_val);    
}



const VLValue &VLValue::operator=(const VLValue &v)
{
    if (&v != this) {
        if (v.m_val != m_val) {
            if (m_val)
                Tcl_DecrRefCount(m_val);
            m_val = v.m_val;
            if (m_val) {
                Tcl_IncrRefCount(m_val);
            }
        }
    }
    
    return *this;
}

VLValue::~VLValue()
{
    if (m_val)
        Tcl_DecrRefCount(m_val);
}
    
Tcl_Obj *VLValue::getObj() const
{
    if (m_val)
        return m_val;
    else
        return Tcl_NewObj();
}

void VLValue::unshare()
{
    if (m_val && Tcl_IsShared(m_val)){
        Tcl_Obj *ob = Tcl_DuplicateObj(m_val);
        Tcl_DecrRefCount(m_val);
        m_val = ob;
        Tcl_IncrRefCount(m_val);
    } 
}

void VLValue::setObj(Tcl_Obj *val)
{
    if (val != m_val) {
        if (m_val)
            Tcl_DecrRefCount(m_val);
        m_val = val;
        if (m_val) {
            Tcl_IncrRefCount(m_val);
        }
    }
}


void VLValue::prepare_modify() 
{
        if (!m_val)
           setObj(Tcl_NewObj());
        else
           unshare();
}
    


void VLValue::setVLValue(Tcl_Obj *val)
{
        setObj(val);
}

    
void *SaveModuleBasicTypes()
{
  /* Save & Restore must be called immediately before or after
   * Initialization. No VLValue-Types must be allocated during 
   * Initialization. */

  return (void*)0;
}

void RestoreModuleBasicTypes(void *)
{
  VLAutoDelete::ReleaseMemory();
  VLProperty::ReleaseMemory();
}

    
void *VLPointer::getPtr() const 
{
    void *l = 0;
    if (m_val)
        TclVP_GetVoidPointerFromObj((Tcl_Interp *)0, m_val, &l);
    return l;
}
    
void VLPointer::setPtr(void *val) 
{
    unshare();
    if (m_val)
        TclVP_SetVoidPointerObj(m_val, val);
    else {
        m_val = TclVP_NewVoidPointerObj(val);
        Tcl_IncrRefCount(m_val);
    }
    
}

VLNodePtr::VLNodePtr(NODEPTR val): VLValue()
{
    setNodePtr(val);
}

VLNodePtr::operator NODEPTR() const
{
    return getNodePtr();
}
    
NODEPTR VLNodePtr::getNodePtr() const 
{
    void *l = 0;
    if (m_val)
        TclVP_GetVoidPointerFromObj((Tcl_Interp *)0, m_val, &l);
    return (NODEPTR)l;
}
    
void VLNodePtr::setNodePtr(NODEPTR val) 
{
    unshare();
    if (m_val)
        TclVP_SetVoidPointerObj(m_val, val);
    else {
        m_val = TclVP_NewVoidPointerObj(val);
        Tcl_IncrRefCount(m_val);
    }
    
}


void (*VLSolverValue::on_change)(VLSolverValue *) = 0;

void VLSolverValue::setVLValue(Tcl_Obj *val)
{
    VLValue::setVLValue(val);
    if (on_change) {
        (*on_change)(this);
    }
}

VLSolverValue::VLSolverValue(): VLValue()
{
}

VLSolverValue::~VLSolverValue()
{
}
VLLong::VLLong(long val): VLValue()
{
    setLong(val);
}

VLLong::operator long() const
{
    return getLong();
}
    
long VLLong::getLong() const 
{
    long l = 0;
    if (m_val)
        Tcl_GetLongFromObj((Tcl_Interp *)0, m_val, &l);
    return l;
}
    
void VLLong::setLong(long val) 
{
    unshare();
    if (m_val)
        Tcl_SetLongObj(m_val, val);
    else {
        m_val = Tcl_NewLongObj(val);
        Tcl_IncrRefCount(m_val);
    }
    
}
VLInt::VLInt(int val): VLSolverValue()
{
    setInt(val);
}

VLInt::operator int() const
{
    return getInt();
}
    
int VLInt::getInt() const 
{
    int val = 0;
    if (m_val)
        Tcl_GetIntFromObj((Tcl_Interp *)0, m_val, &val);
    return val;
}
    
void VLInt::setInt(int val) 
{
    unshare();
    if (m_val)
        Tcl_SetIntObj(m_val, val);
    else {
        m_val = Tcl_NewIntObj(val);
        Tcl_IncrRefCount(m_val);
    }
    
}

char *VLInt::getNewArgs(char *buffer)
{
   //sprintf (buffer, "ptr%x.i", (unsigned int)this);
   sprintf (buffer, "ptr%p.i", this);
   return buffer;
}

char *VLInt::getValArgs(char *buffer)
{
   sprintf (buffer, "%d", getInt());
   return buffer;
}
VLBoolean::VLBoolean(int val): VLValue()
{
    setBoolean(val);
}

VLBoolean::operator int() const
{
    return getBoolean();
}
    
int VLBoolean::getBoolean() const 
{
    int val = 0;
    if (m_val)
        Tcl_GetBooleanFromObj((Tcl_Interp *)0, m_val, &val);
    return val;
}
    
void VLBoolean::setBoolean(int val) 
{
    unshare();
    if (m_val)
        Tcl_SetBooleanObj(m_val, val);
    else {
        m_val = Tcl_NewBooleanObj(val);
        Tcl_IncrRefCount(m_val);
    }
    
}
VLString::VLString(VLCharPtr val): VLValue()
{
    setCharPtr(val);
}

VLString::VLString(int val): VLValue()
{
    m_val = Tcl_NewIntObj(val);
}

VLString::operator VLCharPtr() const
{
    return getCharPtr();
}
    
VLCharPtr VLString::getCharPtr() const 
{
    if (m_val)
        return *new VLAutoDeleteString(Tcl_GetStringFromObj(m_val, (int *)0)); 
    else
        return "";    
}

void VLString::setCharPtr(VLCharPtr val) 
{
    unshare();
    if (m_val)
        Tcl_SetStringObj(m_val, val, -1);
    else {
        m_val = Tcl_NewStringObj(val, -1);
        Tcl_IncrRefCount(m_val);
    }
}


VLString *VLString::append(VLCharPtr val)
{
    prepare_modify();
    Tcl_AppendStringsToObj(m_val,val,NULL);
    return this;
}


VLDouble::VLDouble(double val): VLValue()
{
    setDouble(val);
}

VLDouble::operator double() const
{
    return getDouble();
}

double VLDouble::getDouble() const 
{
    double val = 0;
    if (m_val)
        Tcl_GetDoubleFromObj((Tcl_Interp *)0, m_val, &val);
    return val;
}

void VLDouble::setDouble(double val) 
{
    unshare();
    if (m_val)
        Tcl_SetDoubleObj(m_val, val);
    else {
        m_val = Tcl_NewDoubleObj(val);
        Tcl_IncrRefCount(m_val);
    }

}
VLList::VLList(): VLValue()
{
}


/* by cs */


VLList::VLList(VLCharPtr s): VLValue(Tcl_NewStringObj(s, -1))
{
}



VLList *VLList::append(Tcl_Obj *obj)
{
    prepare_modify();
    Tcl_ListObjAppendElement((Tcl_Interp *)0, m_val, obj);
    return this;
}

VLList *VLList::append(int i)
{
    prepare_modify();

    Tcl_ListObjAppendElement((Tcl_Interp *)0, m_val, Tcl_NewIntObj(i));
    return this;
}

VLList *VLList::append(VLCharPtr s)
{
    prepare_modify();

    Tcl_ListObjAppendElement((Tcl_Interp *)0, m_val, Tcl_NewStringObj(s, -1));
    return this;
}

VLList *VLList::append(double d)
{
    prepare_modify();

    Tcl_ListObjAppendElement((Tcl_Interp *)0, m_val, Tcl_NewDoubleObj(d));
    return this;
}

VLList *VLList::append(const VLValue &v)
{
    return append(v.getObj());
}

VLList *VLList::appendList(const VLList &v)
{
    prepare_modify();
    Tcl_ListObjAppendList((Tcl_Interp *)0, m_val, v.getObj());
    return this;
}


/* by cs: Dies verursacht schwer zu findende Fehler: 
Tcl_ListObjAppendElement((Tcl_Interp *)0, m_val, TclVP_NewVoidPointerObj(ptr));
*/

VLList *VLList::append(VLCharPtr name, int i)
{
    prepare_modify();

    Tcl_ListObjAppendElement((Tcl_Interp *)0, m_val, Tcl_NewStringObj(name, -1));
    Tcl_ListObjAppendElement((Tcl_Interp *)0, m_val, Tcl_NewIntObj(i));
    return this;
}

VLList *VLList::append(VLCharPtr name, VLCharPtr s)
{
    prepare_modify();

    Tcl_ListObjAppendElement((Tcl_Interp *)0, m_val, Tcl_NewStringObj(name, -1));
    Tcl_ListObjAppendElement((Tcl_Interp *)0, m_val, Tcl_NewStringObj(s, -1));
    return this;
}

VLList *VLList::append(VLCharPtr name, double d)
{
    prepare_modify();

    Tcl_ListObjAppendElement((Tcl_Interp *)0, m_val, Tcl_NewStringObj(name, -1));
    Tcl_ListObjAppendElement((Tcl_Interp *)0, m_val, Tcl_NewDoubleObj(d));
    return this;
}

VLList *VLList::append(VLCharPtr name, void *ptr)
{
    prepare_modify();

    Tcl_ListObjAppendElement((Tcl_Interp *)0, m_val, Tcl_NewStringObj(name, -1));
    Tcl_ListObjAppendElement((Tcl_Interp *)0, m_val, TclVP_NewVoidPointerObj(ptr));
    return this;
}

int VLList::getLength() 
{
    int len = 0;

    if (m_val)
        Tcl_ListObjLength((Tcl_Interp *)0, m_val, &len);
    return len;
}

VLString VLList::getIndex(int i)
{
   Tcl_Obj *objPtr;
   if(Tcl_ListObjIndex((Tcl_Interp *)0, m_val, i, &objPtr) == TCL_OK) return VLString(objPtr);
   assert(0);
   return VLString("");
}

VLPointArray::VLPointArray(): VLSolverValue()
{
}

VLPointArray::VLPointArray(int x, int y): VLSolverValue()
{
    setXY(0, x, y);
}

VLPointArray::VLPointArray(int x1, int y1, int x2, int y2): VLSolverValue()
{
    setXY(0, x1, y1);
    setXY(1, x2, y2);
}

int VLPointArray::getX(int index, int defaultval) const
{
    double x = defaultval;
    Tcl_Obj *X = (Tcl_Obj *)0;
    
    if (m_val) {
        if (Tcl_ListObjIndex((Tcl_Interp *)0, m_val, index * 2, &X) == TCL_OK) {
            if (X != (Tcl_Obj *)0) {
                Tcl_GetDoubleFromObj((Tcl_Interp *)0, X, &x);
            }
        }
    }
    
    return int(x);
}

int VLPointArray::getY(int index, int defaultval) const
{
    double y = defaultval;
    Tcl_Obj *Y = (Tcl_Obj *)0;
    
    if (m_val) {
        if (Tcl_ListObjIndex((Tcl_Interp *)0, m_val, index * 2 + 1, &Y) == TCL_OK) {
            if (Y != (Tcl_Obj *)0) {
                Tcl_GetDoubleFromObj((Tcl_Interp *)0, Y, &y);
            }
        }
    }
    
    return int(y);
}
 
int VLPointArray::getNumbPoints() const
{
    int len = 0;
    
    if (m_val)
        Tcl_ListObjLength((Tcl_Interp *)0, m_val, &len);
    return len / 2;
}

void VLPointArray::removePoints(int index, int count)
{
    // Tcl_ListObjReplace strzt ab, wenn index + count > len && count <= len gilt.

    int len;
    if (!m_val) return;
    len = Tcl_ListObjLength((Tcl_Interp *)0, m_val, &len);

    if (2 * index + 2 * count > len) 
        count = len / 2 - index;

    Tcl_ListObjReplace((Tcl_Interp *)0, m_val, 2 * index, 2 * count, 0, (Tcl_Obj **)0);
}

void VLPointArray::setXY(int index, int x, int y)
{
    Tcl_Obj *coords[2];
    int len;
    
    unshare();
    len = getNumbPoints();

    if (index > len) {
        coords[1] = Tcl_NewIntObj(0);
        coords[0] = Tcl_NewIntObj(0);
        while (index > len) {
            if (m_val) 
                Tcl_ListObjReplace((Tcl_Interp *)0, m_val, 2*len, 0, 2, coords);
            else {
                m_val = Tcl_NewListObj(2, coords);
                Tcl_IncrRefCount(m_val);
            }
            
            len++;
        }
    }
    
    coords[0] = Tcl_NewIntObj(x);
    coords[1] = Tcl_NewIntObj(y);


    if (m_val) 
        Tcl_ListObjReplace((Tcl_Interp *)0, m_val, index * 2, (index < len) ? 2 : 0, 2, coords);
    else {
        m_val = Tcl_NewListObj(2, coords);
        Tcl_IncrRefCount(m_val);
    }

}


void VLPointArray::print(char *s="")
  {
  printf("%s\n",s);
  for(int i=0; i<getNumbPoints(); i++)
    {
    printf("(%d,%d)\n",getX(i),getY(i)); 
    }
  }

char *VLPointArray::getNewArgs(char *buffer)
{
   //sprintf(buffer, "ptr%x.x ptr%x.y", (unsigned int)this, (unsigned int)this);
   sprintf(buffer, "ptr%p.x ptr%p.y", this, this);
   return buffer;
}

char *VLPointArray::getValArgs(char *buffer)
{
   sprintf(buffer, "%d %d", getX(), getY());
   return buffer;
}

Tcl_Obj *GetVLProp(VLProperty *p, char *name)
{
  VLList l;
  l.append("::property::getProperty");
  l.append(VLPointer(p->m_data));
  l.append(name);
  return l.eval();
}


Tcl_Obj *GetVLProp(VLProperty *p, char *name, const VLValue &defaultvalue)
{
  VLList l;
  l.append("::property::getProperty");
  l.append(VLPointer(p->m_data));
  l.append(name);
  l.append(defaultvalue);
  return l.eval();
}

void SetVLProp(const VLProperty *p, char *name, const VLValue &value)
{
  VLList l;
  l.append("::property::setProperty");
  l.append(VLPointer(p->m_data));
  l.append(name);
  l.append(value.getObj());
  l.eval();
}


Tcl_Obj *GetVLProp(NODEPTR p, char *name)
{
  VLList l;
  l.append("::property::getProperty");
  l.append(VLPointer(p));
  l.append(name);
  return l.eval();
}


void SetVLProp(NODEPTR p, char *name, const VLValue &value)
{
  VLList l;
  l.append("::property::setProperty");
  l.append(VLPointer(p));
  l.append(name);
  l.append(value.getObj());
  l.eval();
}



VLProperty::vlpropdata *VLProperty::last_vlpropdata = 0;

int VLProperty::delete_me()
{
    VLProperty **start = (VLProperty **)obstack_base(&(m_data->m_obst));
    VLProperty **end = (VLProperty **)obstack_next_free(&(m_data->m_obst));
    VLProperty **base = start;
    
    while (start < end)
    {
        if (*start == this)
        {
            obstack_next_free(&(m_data->m_obst)) = (char *)--end;
            
            if (start != end)
                *start = *end;
            
            if (base == end) 
                /* Komplett leer ! */
                return 1;
            else
                return 0;
        }
        start++;
    }
    
    /* Must not be reached! */
    assert(0);
}

void VLProperty::remove_me(vlpropdata *data)
{
    obstack_free(&(data->m_obst), 0);

    vlpropdata **d = &last_vlpropdata;
    while (*d) {
        if ((*d) == data) {
            *d = (*d)->m_next;
            break;
        }
        else
            d = &((*d)->m_next);
    }

    delete data;

}

void VLProperty::ReleaseMemory()
{
    vlpropdata *d = last_vlpropdata;
    while (d)
    {
        d->m_key = NoKey;
        d = d->m_next;
    }
}

VLProperty::VLProperty():
    m_data(0),
    m_node(NULLNODEPTR)
{
    assert (this != 0)

    m_data = new vlpropdata;
    m_data->m_next = last_vlpropdata;
    last_vlpropdata = m_data;

    obstack_begin(&(m_data->m_obst), 16 * sizeof(void *));
    *((VLProperty **&)obstack_next_free(&(m_data->m_obst)))++ = this;
    m_data->m_key = NoKey;
}

void VLProperty::setNode(NODEPTR my_node)
{
    m_node = my_node;
}


VLProperty::~VLProperty()
{
    if (delete_me())   // Last VLProperty in vlpropdata?
        remove_me(m_data);
}

DefTableKey VLProperty::getDefTableKey()
{
    if (m_data->m_key == NoKey)
        m_data->m_key = NewKey();
    
    return m_data->m_key;
}

void VLProperty::associateWith(VLProperty *vlp)
{
    if (vlp) {
        VLProperty **start = (VLProperty **)obstack_base(&(m_data->m_obst));
        VLProperty **end = (VLProperty **)obstack_next_free(&(m_data->m_obst));

        vlpropdata *data = m_data;            /* Dieser wird gelscht. */

        obstack_make_room(&(vlp->m_data->m_obst), obstack_object_size(&(data->m_obst)));

        // Fr alle VLProperties, die dieses m_data speichern.
        while (start < end)
        {   // Trage Zeiger in vlp->m_data ein.
            *((VLProperty **&)obstack_next_free(&(vlp->m_data->m_obst)))++ = *start;
            // Speichere neuen data-Zeiger.
            (*start)->m_data = vlp->m_data;

            start++;
        }

        // auf data zeigt jetzt niemand mehr, hoffe ich.
        remove_me (data);
    }
}

int VLProperty::isAssociatedWith(VLProperty *vlp)
{
    if (!vlp)
        return 0;

    if (vlp == this)
        return 1;
    
    if (vlp->m_data == m_data)
        return 1;
    
    return 0;
}


void VLProperty::removeAssociations()
{
    if (!delete_me())
    {   // In m_data sind noch andere eingetragen. Es wird ein neues Pgm-Objekt bentigt.
        m_data = new vlpropdata;
        m_data->m_next = last_vlpropdata;
        last_vlpropdata = m_data;
        obstack_begin(&(m_data->m_obst), 16 * sizeof(void *));
    }
    
    obstack_make_room(&(m_data->m_obst), sizeof(void *));
    *((VLProperty **&)obstack_next_free(&(m_data->m_obst)))++ = this;
    m_data->m_key = NoKey;
}

