#ifndef BASICTYPES_H
#define BASICTYPES_H

#include <stddef.h>

#include "deftbl.h"
#include "obstack.h"
#include "eliproto.h"
#include "tcl.h"
#include "nodeptr.h"


/*
Wert, der (falls dynamisch) genau fuer einen
Hauptschleifen-Zyklus des visuellen Editors gilt.  Die Klasse
enthaelt ueberladene Varianten der new und
delete-Operatoren. Zur Unterstuetzung des Aufraeumens vor
einem neuen AG-Lauf werden alle per new allokierten Werte durch den
new-Operator in eine Liste eingereiht.  Delete darf also nicht von
Hand aufgerufen werden! Es wird die Tcl-Speicherverwaltung verwendet.

ANM: Die Autodelete-Informationen machen die Behandlung statischen
Variablen ineffizienter. Prinzipiell
wuerden sie nur bei dynamisch
verwalteten Objektinstanzen benoetigt. Man sollte dafuer
eine Separate Verwaltung realisieren.

*/

class VLAutoDelete
{
private:
    VLAutoDelete *m_next;
    static VLAutoDelete *last_allocated, *all_allocated;

public:
    VLAutoDelete();
    virtual ~VLAutoDelete();

    VLAutoDelete(const VLAutoDelete &v): m_next(0) {} // nicht m_next uebernehmen!
    void operator=(const VLAutoDelete &v) {} // nicht m_next uebernehmen!
    void *operator new(size_t len);
    void operator delete(void *mem);
    static void ReleaseMemory();
};

typedef VLAutoDelete *VLAutoDeletePtr;

                


class VLAutoDeleteString: public VLAutoDelete
{
protected:
    char *m_val;
public:
    VLAutoDeleteString(char *s) { m_val=strdup(s); }
    virtual ~VLAutoDeleteString() { free(m_val); }
    operator char *() { return m_val; }
};


/**
Zur Kommunikation mit dem Editor ist jeder der hier eingefuehrten Typen
eine Spezialisierung von VLValue. Die Klasse VLValue
unterstuetzt die Konversion ihrer Werte in Werte vom Typ Tcl_Obj
mit den Methoden getObj und setObj. Ueber diese Methoden werden
die Terminale des abstrakten Strukturbaums vom generierten
Interfacemodul aus gelesen und geschrieben.

Die Klasse VLValue unterstuetzt weiterhin Kopier- und
Zuweisungsoperator und verwendet dabei das in Tcl gebraeuchliche
Reference Counting. 
*/

class VLValue: public VLAutoDelete
{


protected:
    Tcl_Obj *m_val;

    void unshare();
    void prepare_modify();

public:
    Tcl_Obj *eval(void); /* by cs */
    VLValue();
    VLValue(const VLValue &v);
    VLValue(Tcl_Obj *val);
    const VLValue &operator=(const VLValue &v);
    void setResult(void);
    int isNull(void) { printf("res: %p\n", m_val); return m_val == NULL; }
    virtual ~VLValue();

    virtual Tcl_Obj *getObj() const;
    virtual void setObj(Tcl_Obj *val);
    virtual void setVLValue(Tcl_Obj *val);
};

typedef VLValue *VLValuePtr;
                

#include "tcl_vpt.h"


class VLPointer: public VLValue
{
public:
    VLPointer(Tcl_Obj * obj) : VLValue(obj) {};
    VLPointer(void *val = 0) { setPtr(val); }
    
    void *getPtr() const;
    void setPtr(void *val);
};

typedef VLPointer *VLPointerPtr;
                
#define VLPOINTER(v) (new VLPointer(v))

#include "tcl_vpt.h"


class VLNodePtr: public VLValue
{
public:
    VLNodePtr(Tcl_Obj * obj) : VLValue(obj) {};
    VLNodePtr(NODEPTR val = 0);
    
    operator NODEPTR() const;
    
    NODEPTR getNodePtr() const;
    void setNodePtr(NODEPTR val);
};

typedef VLNodePtr *VLNodePtrPtr;
                
#define VLNODEPTR(v) (new VLNodePtr(v))


/**
  Eine Spezialisierung von VLValue, die ein Update von Variablen durch
  den Constraintsover unterstuetzt.
*/

class VLSolverValue: public VLValue
{
public:
    VLSolverValue(Tcl_Obj * obj) : VLValue(obj) {};
    VLSolverValue();
    ~VLSolverValue();    

    static void (*on_change)(VLSolverValue *);
    void setVLValue(Tcl_Obj *val);

    // buffer points to a buffer with at least 256 chars free space on entry.
    virtual char *getNewArgs(char *buffer) = 0;
    virtual char *getValArgs(char *buffer) = 0;
};

/**
 * Gegenstueck zu long
 */

class VLLong: public VLValue
{
public:
    VLLong(long val = 0);
    
    operator long() const;
    
    long getLong() const;
    void setLong(long val);
};

typedef VLLong *VLLongPtr;
                
#define VLLONG(v) (new VLLong(v))

/**
 * Gegenstueck zu int
 */

class VLInt: public VLSolverValue
{
public:
    VLInt(Tcl_Obj * obj) : VLSolverValue(obj) {};
    VLInt(int val);
    VLInt() {};

    operator int() const;
    
    int getInt() const;
    void setInt(int val);

    virtual char *getNewArgs(char *);
    virtual char *getValArgs(char *);
};

typedef VLInt *VLIntPtr;
                
#define VLINT(v) (new VLInt(v))
#define VLINTVAL(v) ((v)->getInt())

/**
 * Gegenstueck zu bool
 */

class VLBoolean: public VLValue
{
public:
    VLBoolean(int val);
    VLBoolean() {}

    operator int() const;
    
    int getBoolean() const;
    void setBoolean(int val);
    
};

typedef VLBoolean *VLBooleanPtr;
                
#define VLBOOLEAN(v) (new VLBoolean(v))

/**RAW-PUBLIC*/
typedef char *VLCharPtr;
/**END-PUBLIC*/

/**
 * Typ einer Zeichenkette. Es ist ein Typecast zwischen diesem Typ und
 * dem Typ VLCharPtr definiert.
 */

class VLString: public VLValue
{
public:
    VLString(Tcl_Obj * obj) : VLValue(obj) {};
    VLString(const VLValue &v) : VLValue(v) {};
    VLString() {};
    VLString(VLCharPtr val);
    VLString(int val);

    operator VLCharPtr() const;
    
    VLCharPtr getCharPtr() const;
    void setCharPtr(VLCharPtr val);
    VLString *append(VLCharPtr val);
};

typedef VLString *VLStringPtr;
                
#define VLSTRING(s) (new VLString(s))

inline VLCharPtr GetVLCharPtr(VLString *s) { return s->getCharPtr(); }


/**
 * Gegenstueck zu double
 */

class VLDouble: public VLValue
{
public:   
    VLDouble(double val = 0.0);

    operator double() const;
    
    double getDouble() const;
    void setDouble(double val);
};

typedef VLDouble *VLDoublePtr;
                
#define VLDOUBLE(v) (new VLDouble(v))

/**
 * Bei jeder Benutzeraktion im User-Interface koennen statt der "eingebauten" 
 * Reaktion des User-Interfaces eigene Routinen (in Tcl/Tk) aufgerufen werden,
 * die programmierte Aktionen durchfuehren. Diese Routinen koennen mit Hilfe
 * von Listen mit zusaetzlichen Informationen versorgt werden.
 */

class VLList: public VLValue
{
public:
    VLList();
    VLList(VLCharPtr s);
    VLList(VLCharPtr s1, const VLList &v) { append(s1); appendList(v); }
    VLList(VLCharPtr s1, VLCharPtr s2, const VLList &v) { append(s1); append(s2); appendList(v); }
    VLList(VLCharPtr s1, VLValue s2, const VLList &v) { append(s1); append(s2); appendList(v); }
    VLList *append(Tcl_Obj *obj);
    VLList *append(int i);
    VLList *append(VLCharPtr s);
    VLList *append(double d);
    VLList *append(const VLValue &v);
    VLList *append(VLCharPtr name, int i);
    VLList *append(VLCharPtr name, VLCharPtr s);
    VLList *append(VLCharPtr name, double d);
    VLList *append(VLCharPtr name, void *ptr);
    VLList *appendList(const VLList &v);

    int getLength();
    VLString getIndex(int i);

};

typedef VLList *VLListPtr;

/**RAW-PUBLIC*/
#define VLEVAL(l) ((l)->eval())
#define VLLIST() (new VLList())
#define VLLIST1(x) ((new VLList())->append(x))
#define VLLIST2(x,y) (VLLIST1(x)->append(y))
#define VLLIST3(x,y,z) (VLLIST2(x,y)->append(z))
#define VLAPPEND(l,x) ((l)->append(x))
#define VLAPPEND2(l,name,x) ((l)->append(name,x))
#define NOVLLIST (VLListPtr(0))
/**END-PUBLIC*/


/**
Zur Erzeugung visueller Objekte in einem Tcl/Tk-Canvas muessen im
Attributauswerter die Koordinaten der Punkte berechnet und in der
Routine zum Zeichnen eines Objekts uebergeben werden. Weiterhin wird
bei der Plazierung neuer Objekte auf dem Canvas die Koordinaten des
Mauszeigers in ein Attribut POS geschrieben, falls es existiert.
Um diese Position zu verwenden muss im Strukturbaum ein Attribut POS
des Typs VLPointArray deklariert werden.

Um aus der Attributierung auf Methoden zuzugreifen koennen die Makros
SELECT bzw.  PTRSELECT verwendet werden.

Um in einem Attributauswerter einen temporaeren Wert anzulegen oder
einem Attribut zuzuordnen, wird der Typ VLPointArrayPtr verwendet
und einem der Makros initialisiert. Man beachte, dass diese Werte nur
bis zum Anfang des naechsten AG-Laufs existieren.
*/

class VLPointArray: public VLSolverValue
{
public:
    VLPointArray();
    VLPointArray(int x, int y);       // Makes a Point Array consisting of one Point
    VLPointArray(int x1, int y1, int x2, int y2);
    VLPointArray(Tcl_Obj * obj) : VLSolverValue(obj) {};
    
    int getX(int index = 0, int defaultval = -1) const;    // Returns X of Point number index
    int getY(int index = 0, int defaultval = -1) const;    // Returns Y of Point number index.
    int getNumbPoints() const;        // Returns Number of Points.
    
    void removePoints(int index, int count = 1);
                                      // Removes the specified points.
    
    void setXY(int index, int x, int y);
                                      // Changes the Point coordinates or
                                      // inserts at the end. Also inserts null
                                      // elements if required.

    void print(char *);

    virtual char *getNewArgs(char *buffer);
    virtual char *getValArgs(char *buffer);
};

typedef VLPointArray *VLPointArrayPtr;

/**RAW-PUBLIC*/
#define NOVLPOINT ((VLPointArrayPtr)0)
#define VLPOINT(x,y) (new VLPointArray(x,y))
#define VLREGION(x1,y1,x2,y2) (new VLPointArray(x1,y1,x2,y2))
/**END-PUBLIC*/



class VLProperty
{
public:
    struct vlpropdata
    {
        DefTableKey m_key;    /* Zum Speichern von Eigenschaften */
        Obstack m_obst;       /* Zur Verwaltung der Attribute */
        vlpropdata *m_next;   /* Nchstes visuelles Objekt */
    };

    vlpropdata *m_data;       /* by cs: public gemacht (nur temporaer!) */
private:
    NODEPTR m_node;           /* Zum Benachrichtigen bei nderungen */


    int delete_me();         /* Aufrumarbeiten, falls eine Instanz
                               * gelscht wird. Normalerweise veranlat
                               * z.B. durch den Destruktor */
    void remove_me(vlpropdata *data);
                              /* Lscht data und entfernt data von der Liste */
    static vlpropdata *last_vlpropdata;
    
public:
    VLProperty();             /* Erzeugt eine neue Property */
    ~VLProperty();
public:
    DefTableKey getDefTableKey();

    void associateWith(VLProperty *vlp);
                              /* Joins two different program objects */
    void removeAssociations();
                              /* deletes this from associations with others */
    int isAssociatedWith(VLProperty *vlp);
                              /* Checks if an association is defined between the given properties */

    void setNode(NODEPTR);
    static void ReleaseMemory();     /* Speicherbereinigung vor einem neuen AG-Durchlauf */

private:                      /* Kopie wird nicht untersttzt, diese Meth. ex. nicht */
    VLProperty(const VLProperty &vlp);
    const VLProperty &operator=(const VLProperty &vlp);
};

typedef VLProperty *VLPropertyPtr;

#define KEYOF(vlpropptr) ((vlpropptr)->getDefTableKey())
#define NULLVLPropertyPtr ((VLPropertyPtr)0)

Tcl_Obj *GetVLProp(VLProperty *p, char *name);
Tcl_Obj *GetVLProp(VLProperty *p, char *name, const VLValue &defaultvalue);
void SetVLProp(const VLProperty *p, char *name, const VLValue &value);

Tcl_Obj *GetVLProp(NODEPTR p, char *name);
void SetVLProp(NODEPTR p, char *name, const VLValue &value);



#endif
