
# /**
#
# Tck/Tk kennt statdardmaessig die folgenden grafischen Primitive (siehe Tcl/Tk-Referenz):
# [arc], [bitmap], [image], [line], [oval], [polygon], [rectangle], [text] und [window]. 
# Die grafische Primitive lassen sich mittels der Funktion [CreateObject] verwenden, um
# visuelle Darstellungen auf den Canvas zu bringen. Die nachfolgend beschriebene
# Schnittstelle dient der Erweiterung dieser Menge von Grundtypen.
#
# Neue grafische Primitive koennen definiert werden, indem 
# eine Tcl-Funktion mit Praefix [userdefinedPrimitive_] definiert wird, die zwei
# Listen als Parameter verarbeitet, nmlich [coords] und [params]. Die Funktion sollte
# eine Liste der folgenden Form zurueckliefern:
#
# [itemtype1 coords1 params1 itemtype2 coords2 params2 ....]
#
# Als Anwendungsbeispiel fuer die Schnittstelle enthaelt diese Datei drei
# Definitionen fuer neue grafische Primitive, naemlich:
# <ul>
#   <li>[line] (Redefinition), 
#   <li>[roundedrectangle] und 
#   <li>[solid3DRectangle].
# </ul>
#
# */

#########################################
#     Nuetzliche Hilfsfunktionen        #
#########################################


proc vec_sub {a b} {
  return [list \
    [expr [lindex $a 0]-[lindex $b 0]] \
    [expr [lindex $a 1]-[lindex $b 1]]]
}

proc vec_normalize {vector length} {
  set x [lindex $vector 0]
  set y [lindex $vector 1]
  set l [expr sqrt($x*$x+$y*$y)]
  return [list [expr $length*$x/$l] [expr $length*$y/$l]]
}

proc vec_rotateRight90 {vector} {
  set x [lindex $vector 0]
  set y [lindex $vector 1]
  return [list $y [expr -$x]]
}


proc vec_transform {drawing rvec qvec shift} {
  foreach {primtype coords options} $drawing {
    foreach {x y} $coords {
      set x1 [lindex $rvec 0]
      set y1 [lindex $rvec 1]
      set x2 [lindex $qvec 0]
      set y2 [lindex $qvec 1]
      set sx [lindex $shift 0]
      set sy [lindex $shift 1]
      lappend newcoords \
        [expr $sx + $x*$x2 + $y*$x1] \
        [expr $sy + $x*$y2 + $y*$y1]
    }
    lappend newdrawing $primtype $newcoords $options
  }
  return $newdrawing
}


proc handleParams {params allowed} {
  upvar param param
  foreach {tag value} $params {
    if { [lsearch -exact $allowed $tag] == -1 } {
      ::log::log "handleParams: tag $tag is not allowed!"
    } else {
      set param($tag) $value
    }
  }
}



#############################################
#   Erweiterung des "line"-Primitivs        #
#############################################

# Es koennen beliebige Pfeilformen definiert werden
# Koordinatensystem: Die Linie kommt von (0,-unendlich) und endet im Punkt (0,0)
# Dieses System wird defaultmaessig um den Faktor 10 skaliert.
# Datenstruktur: Genau wie der Rueckgabewert von userdefinedPrimitive-Funktionen
# Eine Pfeilform muss den Praefix "lineend_" besitzen!

set lineend_normal      {line    {-0.5 -1 0 0 0.5 -1} {}}
set lineend_inherits    {polygon {0 0 -0.5 -1 0.5 -1} { -fill white -outline black}}
set lineend_aggregation {polygon {0 0 -0.3 -0.5 0 -1 0.3 -0.5} { -fill white -outline black}}
set lineend_composition {polygon {0 0 -0.3 -0.5 0 -1 0.3 -0.5} { -fill black -outline black}}


proc userdefinedPrimitive_line {coords params} {
  set param(-arrowSize1) 10
  set param(-arrowSize2) 10
  set param(-arrowShape1) {}
  set param(-arrowShape2) {}
  handleParams $params "-arrowSize1 -arrowSize2 -arrowShape1 -arrowShape2 -width -fill -stipple \
    -arrow -arrowshape -smooth -dash"

  #Linienende 1
  if { $param(-arrowShape1) == "" } {set arc1 ""} else {
    set last [lrange $coords 0 1]
    set prelast [lrange $coords 2 3]
    set rvec [vec_sub $last $prelast]
    set rvec [vec_normalize $rvec $param(-arrowSize1)]
    set qvec [vec_rotateRight90 $rvec]
    set arc1 [vec_transform [subst "\$::lineend_$param(-arrowShape1)"] $rvec $qvec $last]
  }

  #Linienende 2
  if { $param(-arrowShape2) == "" } {set arc2 ""} else {
    set last [lrange $coords [expr [llength $coords]-2] end]
    set prelast [lrange $coords [expr [llength $coords]-4] [expr [llength $coords]-2]]
    set rvec [vec_sub $last $prelast]
    set rvec [vec_normalize $rvec $param(-arrowSize2)]
    set qvec [vec_rotateRight90 $rvec]
    set arc2 [vec_transform [subst "\$::lineend_$param(-arrowShape2)"] $rvec $qvec $last]
  }

  unset param(-arrowSize1)
  unset param(-arrowSize2)
  unset param(-arrowShape1)
  unset param(-arrowShape2)
  set params [array get param]
  return [concat [list "line" $coords $params] $arc1 $arc2]
}



#############################################
#             Rounded Rectangle             #
#############################################


proc userdefinedPrimitive_roundedrectangle {coords params} {
    set param(-cornerradius) 20
    handleParams $params "-cornerradius -width -fill -outline"

    set x1 [lindex $coords 0]
    set y1 [lindex $coords 1]
    set x2 [lindex $coords 2]
    set y2 [lindex $coords 3]
    set d $param(-cornerradius)

    lappend co $x1 [expr $y1+$d]
    lappend co $x1 $y1  
    lappend co [expr $x1+$d] $y1  

    lappend co [expr $x1+$d] $y1  
    lappend co [expr $x2-$d] $y1  
    lappend co [expr $x2-$d] $y1  

    lappend co $x2  $y1  
    lappend co $x2  [expr $y1+$d]
    lappend co $x2  [expr $y1+$d]

    lappend co $x2  [expr $y2-$d]
    lappend co $x2  [expr $y2-$d]
    lappend co $x2  $y2  

    lappend co [expr $x2-$d] $y2  
    lappend co [expr $x2-$d] $y2  

    lappend co [expr $x1+$d] $y2  
    lappend co [expr $x1+$d] $y2  
    lappend co $x1  $y2  

    lappend co $x1  [expr $y2-$d]
    lappend co $x1  [expr $y2-$d]
    lappend co $x1  [expr $y1+$d]
    
    unset param(-cornerradius)
    set params [array get param]
    return [list "polygon" $co [concat $params "-smooth yes"]]
}  



#############################################
#              solid3DRectangle             #
#############################################


proc userdefinedPrimitive_solid3DRectangle {coords params} {

    set x1 [lindex $coords 0]
    set y1 [lindex $coords 1]
    set x2 [lindex $coords 2]
    set y2 [lindex $coords 3]

    set res ""

    lappend res "rectangle"  "$x1 $y1  $x2 $y2"           "-fill gray"
    lappend res "line"       "$x1 $y2  $x1 $y1  $x2 $y1"  "-fill lightgray -width 2"
    lappend res "line"       "$x2 $y1  $x2 $y2  $x1 $y2"  "-fill darkgray -width 2"
    return $res
}



