# ----------------------------------------------------------------------
# Routinen zur Anzeige von Standarddialogen
# ----------------------------------------------------------------------
# $Id: commondlg.tcl,v 3.6 2000/02/01 12:30:05 mjung Exp $

namespace eval ::CDlg {
}

proc ::CDlg::Init {} {
    global tcl_platform

    # Text to appear as default in the Ok-Button of a ::CDlg::Information
    option add *Info*ok.text "Ok" widgetDefault

    # Text to appear as default in the Ok-Button of a ::CDlg::Error
    option add *Error*ok.text "Ok" widgetDefault

    # Make background white and foreground black for some widgets to
    # make them appear less tcltk-like and more natural.

    option add *Entry.background white widgetDefault
    option add *Entry.highlightbackground blue widgetDefault
    option add *Listbox.background white widgetDefault
    option add *Listbox.highlightbackground blue widgetDefault
    option add *Text.background white widgetDefault
    option add *Text.highlightbackground blue widgetDefault
    if {$tcl_platform(platform) == "unix"} {
	option add *Radiobutton.selectColor black widgetDefault
	option add *Checkbutton.selectColor black widgetDefault
	option add *Menu.selectColor black widgetDefault
    }

    # Raise the dialog if button is pressed anywhere (grab!) on the
    # screen. Requires a matching bindtags list.

    bind CDlgModalDialog <ButtonPress> {
	wm deiconify %W
	raise %W
    }
}

proc ::CDlg::Create {class {title ""} {win "auto"}} {
    # Creates an empty dialog, consisting of two frames, an info-frame
    # and a button-frame separated by a horizontal 3D-line.
    #
    # class      Classname of dialog, also used as window title, if title
    #            is empty.
    #            Under the classname, the user can modify window colors
    #            fonts, etc. by entering settings to his .Xdefaults-file.
    # win        Pathname of the window, optionally computed automagically.

    if {$win == "auto"} {
	set count 0
	set win ".dialog[incr count]"
	while {[winfo exists $win]} {
	    set win ".dialog[incr count]"
	}   
    }

    if {$title == ""} {
	set title $class
    }

    toplevel $win -class $class
    frame $win.info
    pack $win.info -expand yes -fill both -padx 2 -pady 2
    frame $win.sep -height 2 -borderwidth 1 -relief sunken
    pack $win.sep -fill x -pady 4
    frame $win.controls 
    pack $win.controls -fill x -padx 4 -pady 4

    wm title $win $title
    wm group $win .
    
    after idle [format {
	update idletasks
	wm minsize %s [winfo reqwidth %s] [winfo reqheight %s]
    } $win $win $win]
    	
    return $win
    
}

proc ::CDlg::Info {win} {
    # Return pathname of info-pane in a dialog created by CDlgCreate

    return "$win.info"
}

proc ::CDlg::Controls {win} {
    # Return pathname of controls-pane in a dialog created by CDlgCreate

    return "$win.controls"
}

proc ::CDlg::Center {window {parent ""}} {
    # Positions a window in the center of an existing one. Default is
    # Center of the screen.
    if {$parent == ""} {
	set X 0
	set Y 0
	set W [winfo screenwidth .]
	set H [winfo screenheight .]
    } else {
	set W [winfo width $parent]
	set H [winfo height $parent]
	set X [winfo rootx $parent]
	set Y [winfo rooty $parent]
    }

    wm withdraw $window
    update idletasks

    set w [winfo reqwidth $window]
    set h [winfo reqheight $window]
    
    wm geometry $window [format "+%d+%d" \
	    [expr $X + ($W - $w)/2] [expr $Y + ($H - $h)/2] ]
    wm deiconify $window
}

proc ::CDlg::Wait { win {foc ""} } {
    # Waits for User Interaction with $win. Some routine must
    # destroy the window what will cause this procedure to terminate.
    # foc is the first input field in the dialog.

    set prevfocus [focus]
    wm deiconify $win    

    if {[lsearch [bindtags $win] CDlgModalDialog] < 0} {
	bindtags $win [linsert [bindtags $win] 0 CDlgModalDialog]
    }

    set prevgrab [grab current]
    if {$prevgrab != ""} {
	grab release $prevgrab
    }

    tkwait visibility $win
    grab $win
    if {"$foc" != ""} {
	focus $foc
    } else {
	focus $win
    }

    tkwait window $win
    focus $prevfocus
    if {$prevgrab != ""} {
	grab $prevgrab
    }
}
 
proc ::CDlg::NotifyDialog { class title bitmap message } {
    # Internal Routine to build Notification Dialogs. Called from
    # Information, Error, etc. The label of the button is not set and
    # must be assigned using an option.

    set win [Create $class]

    set info [Info $win]
    label $info.icon -bitmap $bitmap
    pack $info.icon -side left -padx 8 -pady 8
    label $info.mesg -text $message -wraplength 4i -justify left
    pack $info.mesg -side right -expand yes -fill both -padx 8 -pady 8
    
    set cntrls [Controls $win]
    button $cntrls.ok -command "\
	    grab release $win\n\
	    destroy $win"
    pack $cntrls.ok -pady 4 -side left -expand yes

    after idle "CDlg::Center $win"

    return $win
}

# Some Ready to use Dialog Procedures.

proc ::CDlg::Information { message {title "Information"}} {
    # A Popup-Window with an i-icon that displays message and has
    # an ok-button.
    
    set win [NotifyDialog "Info" $title info $message]

    Wait $win
}

proc ::CDlg::Error { message {title "Error"}} {

    set win [NotifyDialog "Error" $title error $message]
    wm title $win "$title"

    Wait $win
}

proc ::CDlg::Confirm { message {ok "OK"} {cancel "Cancel"}} {
    # Displays a dialog containing the message. The user chooses between
    # two buttons. The text on the button the user pressed is returned.

    variable result ""

    set CDlg(Confirm) ""

    set win [NotifyDialog "Confirm" "Confirm" questhead $message]
    set cntrls [Controls $win]
    $cntrls.ok configure -text $ok
    $cntrls.ok configure -command [format {
	grab release %s
	destroy %s
	set ::CDlg::result %s
    } $win $win $ok]

    button $cntrls.cancel -text $cancel -command [format {
	grab release %s
	destroy %s
	set ::CDlg::result "%s"
    } $win $win $cancel]
	
    pack $cntrls.cancel -side left -expand yes
        
    Wait $win

    return $result
}

# Some Procedures to facilitate creation of dialog parts.
#
# Routines with a H suffix lay out the label on the left of the entry 
# and require an additional parameter that specifies the width of the label.
# Routines with a V suffix lay out the label over the entry. The label may
# also be empty and will then not be added to the dialog.
#
# name: the name of the new widget to create.
# label: The message to typeset on the left or over the entry.
# args: Additonal arguments supplied to the input widget.
#
# Routines return the widget-name for the input widget (to be used as the
# focus parameter to CDlg::Wait.

# EditEntry is a single line entry widget.
proc ::CDlg::EditEntryH { name label width args } {

    set widgetargs ""
    foreach {option value} $args {
	switch -exact -- $option {
	    -variable {lappend widgetargs -textvariable $value}
	    default {lappend widgetargs $option $value}
	}
    }

    frame $name
    label $name.label -text $label -width $width -anchor w
    eval {entry $name.entry -relief sunken} $widgetargs
    pack $name.label -side left
    pack $name.entry -side right -fill x -expand true
    pack $name -padx 10 -pady 5 -side top -fill x
    return $name.entry
}

proc ::CDlg::EditEntryV { name label args } {
    frame $name

    set widgetargs ""
    foreach {option value} $args {
	switch -exact -- $option {
	    -variable {lappend widgetargs -textvariable $value}
	    default {lappend widgetargs $option $value}
	}
    }

    if {$label != ""} {
	label $name.label -text $label -anchor w
	pack $name.label -side top -expand true -fill x
    }

    eval {entry $name.entry -relief sunken} $widgetargs
    pack $name.entry -side top -fill x -expand true
    pack $name -padx 10 -pady 5 -side top -expand true -fill x
    return $name.entry
}

# EditText is a multi-line text-entry widget using a courier font.
proc ::CDlg::EditTextH { name label width args } {
    set widgetargs ""
    set varname ""
    set textwidth 60
    set textheight 6
    foreach {option value} $args {
	switch -exact -- $option {
	    -variable {set varname $value}
	    -height {set $height $value}
	    -width {set $width $value}
	    default {lappend widgetargs $option $value}
	}
    }

    frame $name
    label $name.label -text $label -width $width -anchor w
    pack $name.label -side left -anchor n

    frame $name.edit
    eval {text $name.edit.text -relief sunken -height $textheight -width $textwidth \
	    -font -*-courier-medium-r-normal-*-12-*-*-*-*-*-*-* \
	    -yscrollcommand [list $name.edit.yscroll set]} $widgetargs
    scrollbar $name.edit.yscroll -orient vertical -command "$name.edit.text yview" \
	    -takefocus 0

    pack $name.edit.text -side left -fill both -expand true
    pack $name.edit.yscroll -side right -fill y

    pack $name.edit -side left -fill both -expand true
    pack $name -padx 10 -pady 5 -side top -expand true -fill both

    $name.edit.text insert end [set $varname]
    
    return $name.edit.text
}

proc ::CDlg::EditTextV { name label args } {
    frame $name

    set widgetargs ""
    foreach {option value} $args {
	switch -exact -- $option {
	    -variable {lappend widgetargs -textvariable $value}
	    default {lappend widgetargs $option $value}
	}
    }

    if {$label != ""} {
	label $name.label -text $label -anchor w -justify left
	pack $name.label -side top -fill x -expand true
    }

    frame $name.edit
    eval {text $name.edit.text -relief sunken \
	    -font -*-courier-medium-r-normal-*-12-*-*-*-*-*-*-*\
	    -yscrollcommand [list $name.edit.yscroll set]} $widgetargs
    scrollbar $name.edit.yscroll -orient vertical -command "$name.edit.text yview" \
	    -takefocus 0

    
    pack $name.edit.text -side left -fill both -expand true
    pack $name.edit.yscroll -side right -fill y
    pack $name.edit -side top -fill both -expand true
    pack $name -padx 10 -pady 5 -side top -expand true -fill both
    return $name.edit.text
}

proc ::CDlg::EditCheckboxH { name label width args } {

    set widgetargs ""
    foreach {option value} $args {
        lappend widgetargs $option $value
    }

    frame $name
    label $name.label -text $label -width $width -anchor w
    eval {checkbutton $name.cb -onvalue 1 -offvalue 0} $widgetargs
    pack $name.label -side left
    pack $name.cb -side left 
    pack $name -padx 10 -pady 5 -side top -fill x
    return $name.cb
}

proc ::CDlg::EditCheckboxV { name label args } {
    frame $name

    set widgetargs ""
    foreach {option value} $args {
        lappend widgetargs $option $value
    }

    if {$label != ""} {
	label $name.label -text $label -anchor w
	pack $name.label -side top -expand true -fill x
    }

    eval {checkbutton $name.cb -onvalue 1 -offvalue 0} $widgetargs
    pack $name.cb -side top 
    pack $name -padx 10 -pady 5 -side top -fill x
    return $name.cb
}

proc ::CDlg::EditRadioH {name label width args} {
    frame $name

    set widgetargs ""
    set cols 2
    foreach {option value} $args {
	switch -exact -- $option {
	    -values {set buttonlist $value}
	    -cols {set cols $value}
	    default {lappend widgetargs $option $value}
	}
    }


    label $name.label -text $label -width $width -anchor w
    pack $name.label -side left -anchor n

    # Name of first frame.
    set fno 0
    set fname $name.ed[incr fno]
    frame $fname
    pack $fname -side left -anchor n -fill x -expand true
    
    set fmax [expr ([llength $buttonlist] / 2) / $cols]
    if {([llength $buttonlist] / 2) % $cols != 0} {
	incr fmax
    }

    set no 0
    foreach {label value} $buttonlist {
	set thisno [incr no]
	if {$no > $fmax} {
	    set fname $name.ed[incr fno]
	    frame $fname
	    pack $fname -side left -anchor n -fill x -expand true
	    set no 0
	    set thisno [incr no]
	}
	eval {radiobutton $fname.r$thisno -anchor w -value $value \
		-text $label} $widgetargs
	pack $fname.r$thisno -side top -fill x
    }
    pack $name -padx 10 -pady 5 -side top -fill x 
    return $name.ed1.r1
}

proc ::CDlg::EditRadioV {name label args} {
    frame $name

    set widgetargs ""
    foreach {option value} $args {
	switch -exact -- $option {
	    -values {set buttonlist $value}
	    default {lappend widgetargs $option $value}
	}
    }

    label $name.label -text $label -anchor w
    pack $name.label -side top -expand true -fill x

    # Name of first frame.
    set fno 0
    set fname $name.ed[incr fno]
    frame $fname
    pack $fname -side left -anchor n -fill x -expand true
    
    set fmax [expr ([llength $buttonlist] / 2) / $cols]
    if {([llength $buttonlist] / 2) % $cols != 0} {
	incr $fmax
    }

    set no 0
    foreach {label value} $buttonlist {
	set thisno [incr no]
	if {$no > $fmax} {
	    set fname $name.ed[incr fno]
	    frame $fname
	    pack $fname -side left -anchor n -fill x -expand true
	    set no 0
	    set thisno [incr no]
	}
	radiobutton $fname.r$thisno -anchor w -value $value -text $label 
	pack $fname.r$thisno -side top -expand true -fill x
    }
    pack $name.edit -side top
    pack $name -padx 10 -pady 5 -side top 
    return $name.edit.r0
}

proc ::CDlg::EditOptionH {name label width args} {
    frame $name

    set widgetargs ""
    foreach {option value} $args {
	switch -exact -- $option {
	    -values {set optionlist $value}
	    -variable {
		set varname $value
	    }
	    default {lappend widgetargs $option $value}
	}
    }

    if {$varname == ""} {
	error "Option -variable not specified"
    }
    upvar #0 $varname variable
    
    label $name.label -text $label -width $width -anchor w
    pack $name.label -side left

    eval {menubutton $name.mb -indicatoron true -menu $name.mb.menu \
	    -relief raised}
    pack $name.mb -side right -expand true -fill x

    menu $name.mb.menu -tearoff 0 
    foreach {label value} $optionlist {
	eval {$name.mb.menu add command -label $label -command "\
		$name.mb configure -text \"$label\"
		set $varname $value"} $widgetargs
	if {"$variable" == "$value"} {
	    $name.mb configure -text $label
	}
    }

    pack $name -padx 10 -pady 5 -side top -fill x
    return $name.mb
}

proc ::CDlg::EditOptionV {name label args} {
    frame $name

    set widgetargs ""
    foreach {option value} $args {
	switch -exact -- $option {
	    -values {set optionlist $value}
	    -variable {
		set varname $value
		lappend widgetargs $option $value
	    }
	    default {lappend widgetargs $option $value}
	}
    }
    if {$varname == ""} {
	error "Option -variable not specified"
    }
    upvar #0 $varname variable
    
    label $name.label -text $label -anchor w
    pack $name.label -side top -expand true -fill x

    eval {menubutton $name.mb -indicatoron true -menu $name.mb.menu \
	    -relief raised}
    pack $name.mb -side top -expand true -fill x

    menu $name.mb.menu -tearoff 0
    foreach {label value} $optionlist {
	eval {$name.mb.menu add radiobutton -label $label -value $value \
		-command "$name.mb configure -text \{$label\}"} $args
	if {"$variable" == "$value"} {
	    $name.mb configure -text $label
	}
    }

    pack $name -padx 10 -pady 5 -side top -expand true -fill x
    return $name.mb
}

# Scale is a numerical scale between two values
proc ::CDlg::EditScaleH { name label width args } {

    set widgetargs ""
    foreach {option value} $args {
	switch -exact -- $option {
	    default {lappend widgetargs $option $value}
	}
    }

    frame $name
    label $name.label -text $label -width $width -anchor w
    eval {scale $name.scale -relief sunken -orient horizontal} $widgetargs
    pack $name.label -side left
    pack $name.scale -side right -fill x -expand true
    pack $name -padx 10 -pady 5 -side top -fill x
    return $name.scale
}

proc ::CDlg::EditScaleV { name label args } {
    frame $name

    set widgetargs ""
    foreach {option value} $args {
	switch -exact -- $option {
	    default {lappend widgetargs $option $value}
	}
    }

    if {$label != ""} {
	label $name.label -text $label -anchor w
	pack $name.label -side top -expand true -fill x
    }

    eval {scale $name.entry -relief sunken -orient horizontal} $widgetargs
    pack $name.entry -side top -fill x -expand true
    pack $name -padx 10 -pady 5 -side top -expand true -fill x
    return $name.entry
}

proc ::CDlg::EditListboxH { name label width args } {
    set lbheight 5
    set varname ""
    set widgetargs ""
    foreach {option value} $args {
	switch -exact -- $option {
	    -height {set lbheight $value}
	    -values {set optionlist $value}
	    -variable {set varname $value}
	    default {lappend widgetargs $option $value}
	}
    }

    # Create the widgets
    frame $name
    label $name.label -text $label -width $width -anchor w
    pack $name.label -side left -anchor n

    frame $name.edit
    eval {listbox $name.edit.lb -relief sunken -height $lbheight -selectmode single \
	    -yscrollcommand [list $name.edit.yscroll set]} $widgetargs
    scrollbar $name.edit.yscroll -orient vertical -command "$name.edit.lb yview" \
	    -takefocus 0

    pack $name.edit.lb -side left -fill both -expand true
    pack $name.edit.yscroll -side right -fill y

    pack $name.edit -side left -fill both -expand true
    pack $name -padx 10 -pady 5 -side top -expand true -fill both

    # Create a new global variable in ::CDlg namespace
    set counter 0
    while {1} {
	set var "::CDlg::listboxdata#[incr counter]"
	upvar #0 $var data
	if {![info exists data]} {
	    break
	}
    }
    bind $name <Destroy> "if \{\"%W\" == \"$name\"\} \{::CDlg::EditListboxDestroy $var\}"

    # Enter names of possible data into the listbox
    set ind 0
    set selid 0
    foreach {description id} "$optionlist" {
	$name.edit.lb insert $ind $description
	set data($ind) $id
	if {"$id" == [set $varname]} {
	    set selid $ind
	}
	incr ind
    }
    $name.edit.lb selection set $selid
    $name.edit.lb see $selid
    bind $name.edit.lb <ButtonPress-1> [format {
	upvar #0 %s ::CDlg::listboxvar 
	set ::CDlg::listboxsel [%s.edit.lb nearest %%y]
	if {$::CDlg::listboxsel >= 0} {
	    set %s $::CDlg::listboxvar($::CDlg::listboxsel)
	}
    } $var $name $varname]
    return $name.edit.lb
}
    
proc ::CDlg::EditListboxDestroy {var} {
    upvar #0 $var data
    if {[info exists data]} {
	unset data
    }
}

proc ::CDlg::EditColorH { name label width args } {
    set widgetargs ""
    set varname ""
    foreach {option value} $args {
	switch -exact -- $option {
	    -variable {set varname $value}
	    default {lappend widgetargs $option $value}
	}
    }

    if {[set $varname] == ""} {
	set $varname "#000000"
    }

    frame $name
    label $name.label -text $label -width $width -anchor w
    eval {menubutton $name.mb \
	    -background [set $varname] \
	    -text "  " \
	    -menu $name.mb.menu \
	    -indicatoron 1 \
	    -direction flush \
	    -activebackground [set $varname]} $widgetargs
    menu $name.mb.menu -tearoff 0
    set count -1
    foreach col \
	    [list "#000000" "#000080" "#008000" "#800000" \
	    "#008080" "#800080" "#808000" "#808080" \
	    "#0000ff" "#00ff00" "#ff0000" "#00ffff" \
	    "#ff00ff" "#ffff00" "#ffffff"] {
	if {[incr count] == 4} {
	    set colbreak 1
	    set count 0
	} else {
	    set colbreak 0
	}

	$name.mb.menu add command -label "   " \
		-background $col \
		-command "::CDlg::EditColorValue $varname $name $col" \
		-columnbreak $colbreak \
		-hidemargin 1 \
		-font "Courier -10" \
		-activebackground $col
    }
    $name.mb.menu add command \
	    -label "..." \
	    -command "::CDlg::EditColorValue $varname $name" \
	    -font "Courier -10" \
	    -hidemargin 1

    pack $name.label -side left
    pack $name.mb -side right -fill x -expand true
    pack $name -padx 10 -pady 5 -side top  -fill x
    return $name.mb
}

proc ::CDlg::EditColorValue {varname name {newcolor ""}} {
    if {$newcolor == ""} {
	set newcolor [tk_chooseColor -initialcolor [set $varname] -parent $name]
    }
    if {$newcolor != ""} {
	$name.mb configure -background $newcolor -activebackground $newcolor
	set $varname $newcolor
    }
}
