# ----------------------------------------------------------------------
# Module Visual
# ----------------------------------------------------------------------
# Procedures to visualize and edit the visualized data structure
#
# $Id: visual.tcl,v 3.44 2001/01/30 11:05:59 cschmidt Exp $

# Create our namespace.
namespace eval ::visual {
}

proc ::visual::tags_setproperty {tagsvar name value} {
  upvar $tagsvar tags
  set value [join $value ","]
  lappend tags "$name,$value"
}

proc ::visual::tags_getproperty {tags name} {
  set propind [lsearch -glob $tags "$name,*"]
  if {$propind >= 0} {
    set value [string range [lindex $tags $propind] [expr [string length $name]+1] end]
    set value [split $value ","]
    return $value
  }
  return ""
}


proc ::visual::updateTitle {win} {
    variable visInfo
    set fileid $visInfo($win-fileid)
    set title "$visInfo($win-title) ([file::modifiedMarker $fileid][::file::getFilename $fileid])"
    ::MDIFrame::SetTitle [::MDIFrame::GetClient $win] $title
}


proc ::visual::init {} {
    variable visInfo
    variable update_called
    variable updatewindows

    # Is update in progress
    set update_called 0
    set updatewindows ""

    set visInfo(windows) ""
    ::event::add_listener tree_changed "::visual::on_update"
    ::event::add_listener tree_delete "::visual::on_delete"

    # Load Images for command bar
    set visInfo(cut) $::cfg::img(button_cut)
    set visInfo(copy) $::cfg::img(button_copy)
    set visInfo(paste) $::cfg::img(button_paste)
    set visInfo(down) $::cfg::img(button_down)
    set visInfo(properties) $::cfg::img(button_properties)

    ::visual::inspoint_init
    ::visual::insert_init
    ::visual::solver_init
    ::visual::move_init
}

# Creates a visual workspace
# win:		Prefix of widget-name in which to create the visual workspace
# rootid:       Id of a root coordinate for the workspace
# attreval:	Name of attribute evaluator that visualizes the program part

proc ::visual::create {win rootid attreval args} {
    # puts "Creating $win"
    variable visInfo

    set selectnode ""
    foreach {opt val} $args {
	switch -- $opt {
	    -selectnode {set selectnode $val}
	}
    }

    # Main Frame
    frame $win -class Visual
    pack $win -expand true -fill both

    frame $win.cmd -relief sunken -borderwidth 2
    pack $win.cmd -side top -fill x
    button $win.cmd.cut -image $visInfo(cut) -state disabled \
	    -command "::visual::cutselection $win 1"
    button $win.cmd.copy -image $visInfo(copy) -state disabled
    button $win.cmd.paste -image $visInfo(paste) -state disabled
    frame $win.cmd.zoom -relief sunken -borderwidth 1
    label $win.cmd.zoom.lab -text "100%" -background white -width 5 -borderwidth 0
    button $win.cmd.zoom.select -image $visInfo(down) -relief raised
    pack $win.cmd.zoom.lab $win.cmd.zoom.select -side left -fill y
    pack $win.cmd.cut $win.cmd.copy $win.cmd.paste -side left -padx 0 -pady 1 -fill y
    menu $win.cmd.zoom.win -tearoff 0 
    set maxgridsize [lindex $::cfg::gridsizes end]
    foreach gridsize $::cfg::gridsizes {
	$win.cmd.zoom.win add command -label "[expr int($gridsize * 100 / $maxgridsize)]%" \
		-command "::visual::setgrid $win $gridsize \"[expr int($gridsize * 100 / $maxgridsize)]%\""
    }
    if {[llength $::cfg::gridsizes] > 1} {
	pack $win.cmd.zoom -padx 3 -pady 1 -fill y -side left
	set nextpad 1
    } else {
	set nextpad 3
    }
    button $win.cmd.properties -image $visInfo(properties) -state disabled \
	    -command "::visual::editselection $win"
    pack $win.cmd.properties -padx $nextpad -pady 1 -fill y -side left

    #button $win.cmd.solver -text "S" -command "::visual::solver_show $win"
    #pack $win.cmd.solver -padx $nextpad -pady 1 -fill y -side left

    bind $win.cmd.zoom.select <1> "tk_popup $win.cmd.zoom.win \[winfo rootx $win.cmd.zoom\] \[expr \[winfo rooty $win.cmd.zoom\] + \[winfo height $win.cmd.zoom\]\]"
    bind $win.cmd.zoom.lab <1> "tk_popup $win.cmd.zoom.win \[winfo rootx $win.cmd.zoom\] \[expr \[winfo rooty $win.cmd.zoom\] + \[winfo height $win.cmd.zoom\]\]"

    # Toolbar
    frame $win.tbar -relief sunken -borderwidth 2
    pack $win.tbar -side left -fill y
    menu $win.contextmenu -tearoff 0

    # Workspace with scrollbars
    frame $win.wrk
    canvas $win.wrk.can -background white -width 300 -height 300 \
	    -yscrollcommand "$win.wrk.ys set" \
	    -xscrollcommand "$win.wrk.xs set" \
	    -cursor crosshair -confine true \
	    -scrollregion { -1 -1 300 300 }
    scrollbar $win.wrk.ys -orient vertical -command "$win.wrk.can yview"
    frame $win.wrk.bt
    scrollbar $win.wrk.xs -orient horizontal -command "$win.wrk.can xview"
    set pad [expr [$win.wrk.ys cget -width] + 2 * \
	    ([$win.wrk.ys cget -bd] + [$win.wrk.ys cget -highlightthickness])]
    frame $win.wrk.pad -width $pad -height $pad
    pack $win.wrk.bt -side bottom -fill x
    pack $win.wrk.pad -in $win.wrk.bt -side right
    pack $win.wrk.xs -in $win.wrk.bt -side bottom -fill x
    pack $win.wrk.ys -side right -fill y
    pack $win.wrk.can -side right -fill both -expand true
    pack $win.wrk -side right -fill both -expand true
    
    # Create dataspace

    set counter 0
    while {1} {
	set var "visData#[incr counter]"
	upvar #0 ::visual::$var data
	if {![info exists data]} {
	    break
	}
    }
    set visInfo($win-data) $var
    set visInfo($win-title) [::MDIFrame::GetTitle [::MDIFrame::GetClient $win]]
    set fileid [::file::getFileId [::vl::getRoot $rootid]]
    set visInfo($win-fileid) $fileid

    lappend visInfo(windows) $win

    # Initialize the data
    set data(workspace) $win.wrk.can

    # Set Client path for Inter-Window drag&drop
    ::MDIFrame::SetClientpath [::MDIFrame::GetClient $win] $data(workspace)
    ::MDIFrame::SetDropcommand [::MDIFrame::GetClient $win] "::visual::accept_interwindowdrop $win"

    # Debug-Code:
    $data(workspace) bind all <2> "::MDIFrame::SetStatus \[$data(workspace) itemcget current -tags\]"

    set data(toolbar) $win.tbar
    set data(rootid) $rootid
    
    $data(workspace) xview moveto 0
    $data(workspace) yview moveto 0

    # Remember name of attribute evaluator
    set data(attreval) $attreval

    # by cs: Einfachste moeglichkeit, um an die id des aktuellen Fensters zu kommen
    set data(windowid) $win

    # Id's der Canvasobjekte, die derzeit / beim letzten update auf dem 
    # Bildschirm sind / waren.
    set data(currentids) ""
    set data(previds) ""

    # Tiefe existierender Canvasobjekte sowie aktuelle Default-Tiefe
    set data(depths) ""
    
    # data(update) == 1 falls grade die Daten auf dem Canvas aktualisiert werden.
    set data(update) 0

    # Alle visuellen Mae sind ein Vielfaches der gridsize, damit
    # Skalierung problemlos mglich ist. 
    setgrid $win $::cfg::defaultgridsize "[expr int($::cfg::defaultgridsize * 100 / $maxgridsize)]%" 0

    # Cleanup
    bind $win <Destroy> "if \{\"%W\" == \"$win\"\} \{::visual::on_destroy $win\}"
    bind $data(workspace) <Configure> "::visual::configurevisual $win %x %y %w %h"


    #::MDIFrame::SetClientcommand [::MDIFrame::GetClient $win] \
    #    "::visual::receivewindowcommand $win"


    ::oop::setClass $win visual
    ::oop::setAlias $win [::MDIFrame::GetClient $win]
    ::file::addWindow $fileid $win

    ::visual::solver_create $win
    ::visual::insert_create $win
    ::visual::move_create $win
    ::visual::drag_create $win

    after idle "
	update idletasks
	::visual::select $win $selectnode
    "

    updateTitle $win
    return "::visual::$var"
}

proc ::visual::wincmd_showSolverInteraction {win} {
    solver_show $win
}

proc ::visual::wincmd_save {win} {
    variable visInfo
    ::file::save $visInfo($win-fileid)
}

proc ::visual::wincmd_saveAs {win} {
    variable visInfo
    ::file::saveAs $visInfo($win-fileid)
}

proc ::visual::wincmd_showErrWindow {win} {
    variable visInfo
    ::file::openErrorList $visInfo($win-fileid)
}

proc ::visual::wincmd_showAbstree {win} {
    variable visInfo
    ::file::activateMainView $visInfo($win-fileid)
}

proc ::visual::wincmd_print {win} {
    variable visInfo
    set var $visInfo($win-data)
    upvar #0 ::visual::$var data

    regexp {[^.]*} [::file::getFilename $visInfo($win-fileid)] rootname
    set wintitle $visInfo($win-title)
    regsub -all " " $wintitle "" wintitle
    regsub -all "\\." $wintitle "_" wintitle
    regsub -all "/" $wintitle "" wintitle

    set filename "$rootname-$wintitle.eps"

    set bbox [$data(workspace) bbox all]
    $data(workspace) postscript -x [lindex $bbox 0] -y [lindex $bbox 1] \
	    -width [expr [lindex $bbox 2] - [lindex $bbox 0]] \
	    -height [expr [lindex $bbox 3] - [lindex $bbox 1]] \
	    -file $filename
    ::CDlg::Information "View saved as '$filename"
}

proc ::visual::wincmd_process {win processor} {
    variable visInfo
    ::file::process $visInfo($win-fileid) $processor
}

	
proc ::visual::setInsertContext {function args} {
    # Known insert Contexts:
    # 
    # toolbar       Applied production from toolbar, no arguments
    # move          Moved node
    #               args[0]: action-code provided to insertion-point
    # dragdrop      Applied production resulting from a drag&drop-operation
    #               from another window. 
    #               args[0]: Id of Node from other window.
    # objdrag       Drag&Drop in a Window resulting in Attribute settings.
    #               args[0]: Argumente der Drag-Stelle
    #               args[1]: Argumente der Drop-Stelle
    #               args[2]: Bounding-Box der Drop-Stelle Liste mit x1, y1, x2, y2
    #               args[3]: Liste mit X/Y-Koordinaten des Maus-Cursors.
    #               args[4]: Partname der Drag-Stelle
    #               args[5]: Partname der Drop-Stelle

    variable insertcontext

    set insertcontext(function) $function
    set insertcontext(args) $args
}

proc ::visual::getInsertContext {} {
    variable insertcontext

    return $insertcontext(function)
}

proc ::visual::getInsertContextArgs {} {
    variable insertcontext

    return $insertcontext(args)
}

proc ::visual::setgrid {win gridsize percenttext {doupdate 1}} {
    variable visInfo
    set var $visInfo($win-data)
    upvar #0 ::visual::$var data

    set data(gridsize) $gridsize
    set data(gridsize2) [expr $data(gridsize) * 2]
    set data(gridsize3) [expr $data(gridsize) * 3]
    set data(gridsize4) [expr $data(gridsize) * 4]
    set data(gridsize5) [expr $data(gridsize) * 5]
    set data(gridsize6) [expr $data(gridsize) * 6]
    set data(gridsize_2) [expr $data(gridsize) / 2]
    set data(gridsize_4) [expr $data(gridsize) / 4]

    $win.cmd.zoom.lab con -text $percenttext
    if {$doupdate} {
	update_after_idle $win 1
    }
} 

proc ::visual::configurevisual {win x y w h} {
    variable visInfo
    set var $visInfo($win-data)
    upvar #0 ::visual::$var data

    set sr [$data(workspace) cget -scrollregion]
    set W [expr [lindex $sr 2] - [lindex $sr 0]]
    set H [expr [lindex $sr 3] - [lindex $sr 1]]
    set newW $W
    set newH $H
    set newX [lindex $sr 0]
    set newY [lindex $sr 1]
    
    if {[expr $w - $x] > $W} {
	set newW [expr $w - $x]
    }
    if {[expr $h - $y] > $H} {
	set newH [expr $h - $y]
    }
    if {$x < $newX} {
	set newX $x
    } 
    if {$y < $newY} {
	set newY $y
    }
    
    if {$newW != $W || $newH != $H} {
	$data(workspace) configure -scrollregion [list $newX $newY $newW $newH]
    }
}


proc ::visual::on_destroy {win} {
    variable visInfo
    if {[info exists visInfo($win-data)]} {
	set var $visInfo($win-data)
	
	# Stop Constraint-Solver it running
	::visual::solver_stop ::visual::$var

	upvar #0 ::visual::$var data
	set windex [lsearch -exact $visInfo(windows) $win]
	set visInfo(windows) [lreplace $visInfo(windows) $windex $windex]
	unset visInfo($win-data)
    }

    # Remove names from toolbar buttons
    foreach var [array names visInfo tbwin-$win.*] {
	unset visInfo($var)
    }
    ::file::forgetWindow $visInfo($win-fileid) $win
}

proc ::visual::on_treeModified {id node} {
    # dummy method (for now); called by file automatically
}

proc ::visual::on_update {id} {
    variable visInfo

    foreach win $visInfo(windows) {
	on_update_win $win $id
    }
}

proc ::visual::on_delete {id} {
    variable visInfo

    foreach win $visInfo(windows) {
	on_delete_win $win $id
    }
}

proc ::visual::on_update_win {win id} {
    variable visInfo
    set var $visInfo($win-data)
    upvar #0 ::visual::$var data

    # by cs: Nachfolgendes stand schon mal so aehnlich da

    # See if our root is there.
    if {![::treeutil::exists $data(rootid)]} {
        # It's gone.
        closeWindow
        return
    }

    if {[::vl::isAncestor $data(rootid) $id]} {
	update_after_idle $win
    }
}

proc ::visual::closeWindow {win} {
    ::MDIFrame::CloseClient [::MDIFrame::GetClient $win] 1
}

proc ::visual::on_delete_win {win id} {
    variable visInfo
    set var $visInfo($win-data)
    upvar #0 ::visual::$var data

    if {$id == [::vl::getRoot $data(rootid)]} {
	closeWindow
	return
    }
    #folgendes ist nicht mehr notwendig (glaube ich)
    #
    #if {$id == $data(selectedid)} {
    #	::visual::select $win
    #}
}

proc ::visual::update_after_idle {win} {
    variable visInfo
    variable updatewindows
    variable update_called
    set var $visInfo($win-data)
    upvar #0 ::visual::$var data

    if {[lsearch -exact $updatewindows $win] < 0} {
	lappend updatewindows $win
    }

    if {$update_called == 0} {
	set update_called 1
	after idle ::visual::update_necessary
    }
}

proc ::visual::update_necessary {} {
    variable updatewindows
    variable update_called

    set wins $updatewindows
    set updatewindows ""

    foreach win $wins {
	::visual::update $win
    }

    set update_called 0
    if {$updatewindows != ""} {
	set update_called 1
	after idle ::visual::update_necessary
    }	
}

# Update visual information. Prevent recursive calls and
# catch errors.

proc ::visual::update {win} {
    variable visInfo
    set var $visInfo($win-data)
    upvar #0 ::visual::$var data


#    puts "update called for a Visual Window"

    if {$data(update)} {
	::log::log "::visual::update called recursively"
#	error "::visual::update called recursively"
	return
    }
    set data(update) 1

    set prevcursor [$data(workspace) cget -cursor]
    $data(workspace) configure -cursor watch

    set ::solvertime 0
    set ::canvastime 0
    set ::solver_calctime 0
    set ::attreval_time 0
    set ::updatetime [time {
    if {[catch {do_update $win} erg]} {
	global errorInfo
	
	log::log "Error during ::visual::update: $erg"
	log::log $errorInfo
	bgerror "Error during ::visual::update: $erg"
    }

    set data(update) 0

    # Redraw the selection
    if { $data(selectedid) != "" } {
        if ![::treeutil::exists $data(selectedid)] {
            # by cs: selektierter Knoten wurde geloescht, also Selektion aufheben
            set data(selectedid) ""
        }
        ::visual::select $win $data(selectedid)
    }
    }]

    set ::updatetime [lindex $::updatetime 0]
    log::info timing "Update: [expr $::attreval_time - $::canvastime - $::solver_calctime], Solver: $::solvertime, Constraints: [expr $::solver_calctime - $::solvertime] Canvas: [expr $::canvastime + $::updatetime - $::attreval_time], Total $::updatetime"

    $data(workspace) configure -cursor $prevcursor
}

proc ::visual::do_update {win} {
    variable visInfo
    set var $visInfo($win-data)
    upvar #0 ::visual::$var data

    # Setze Liste der Id's zurck.

    set data(previds) $data(currentids)
    set data(currentids) ""
    set data(depths) ""

    # Vergesse die zustzlich gesetzten Arguments
    foreach index [array names data vargs,*] {
	unset data($index)
    }

    # Vergesse die Contextmen-Eintrge
    foreach index [array names data contextmenu,*] {
	unset data($index)
    }

    foreach index [array names data additionalEditContext,*] {
	unset data($index)
    }

    # Eigentliches Malen.

    update_id ::visual::$var

    # Lsche alle nicht mehr gebrauchten visuellen Elemente.

    set deleted 0
    foreach workspaceId $data(previds) {
	if {[lsearch -exact $data(currentids) $workspaceId] == -1} {
	    $data(workspace) delete $workspaceId
	    set deleted 1
	}
    }

    # if {$deleted} {::balloon::cancel}

    # Schaue nach ob die Gre des Workspace ausreicht. Falls nein, mach ihn
    # grer.

    set sr [$data(workspace) cget -scrollregion]
    set wrkWidth [lindex $sr 2]
    set wrkHeight [lindex $sr 3]

    set req [$data(workspace) bbox all]
    set reqWidth [lindex $req 2]
    set reqHeight [lindex $req 3]

    if {$reqWidth > $wrkWidth}    { set wrkWidth $reqWidth }
    if {$reqHeight > $wrkHeight}  { set wrkHeight $reqHeight }

    $data(workspace) configure -scrollregion [list -1 -1 $wrkWidth $wrkHeight]

    # Bringe die Canvasobjekte in die richtige Reihenfolge
    set lastdepth 0
    foreach depth [lsort -integer $data(depths)] {
	if {$lastdepth == 0} {
	    $data(workspace) lower "depth_$depth"
	} else {
	    $data(workspace) raise "depth_$depth" "depth_$lastdepth"
	}
	set lastdepth $depth
    }
    catch {$data(workspace) raise inspoint}
    catch {$data(workspace) raise drag}

    # prfe, ob zu den Toolbarbuttons auch Einfgestellen existieren
    foreach index [array names data toolbutton,*] {
	set kind $data($index)
	set win [string range $index 11 end]
	
	set state "disabled"
	foreach code "fi fb mi mb" {
	    if {[$data(workspace) find withtag inspoint$code,$kind] != ""} {
		set state "normal"
		break
	    }
	}	    

	$win configure -state $state
    }
}

proc ::visual::update_id {varname} {
    upvar #0 $varname data

    set ::visual::currVarname $varname
    ::property::preOtherRun

    ::log::info execution "Executing view evaluator $data(attreval)"
    set ::attreval_time [lindex [time {
	::vl::run $data(attreval) $data(rootid)
    }] 0]
    ::log::info execution "View evaluator $data(attreval) ready!"
    ::property::postOtherRun

}


# A Procedure that creates or updates a visual object.
# Parameters:
# id         the abstract tree node id that is visualized through this object
# partname   together with id, this is used to build a unique name for the
#            created object
# type       some kind of visual object. Must be acceptable for the 'canvas create'
#            command.
# depth      depth of object. Objects with smaller depth hide objects with higher depth
#            Within the same depth, the order of creation is used.
# coords     Coordinates for the object. Number of coords passed here depend of 'type'.
# args       Additional attributes passed to canvas create or canvas itemconfigure command.

proc ::visual::check_partname {partname} {
    upvar #0 $::visual::currVarname data

    if {[string first " " $partname] >= 0} {
	::log::log "Partname must not contain spaces"
	return [join [list $partname] "+"]
    } else {
	return $partname
    }
}

proc ::visual::create_object {id partname type depth coords args} {
  set time [time {
  if { [info commands "userdefinedPrimitive_$type"] != "" } {
    set drawing [userdefinedPrimitive_$type $coords $args]
  } else {
    set drawing [list $type $coords $args]
  }
  set counter 1
  foreach {type coords args} $drawing {
    set objid [create_primitive_object $id "$partname$counter" $type $depth $coords $args]
    incr counter
  }
  }] 
  set ::canvastime [expr $::canvastime + [lindex $time 0]]
  
  return $objid
}

proc ::visual::create_primitive_object {id partname type depth coords args2} {
    upvar #0 $::visual::currVarname data
    set args $args2
    set rulecode [::vl::getCode $id]
    set symbcode [::vl::lhs $rulecode]
    set canvas_id "node#$id"
    set canvas_uid "node,$id,[check_partname $partname]"
    set canvas_rule "rule_$rulecode"
    set canvas_symb "symb_$symbcode"
    set canvas_depth "depth_$depth"
    set canvargs $args

#    puts "Creating Object $id $partname $type $depth $coords $args"

    if {$type == "text"} {
	# Use default-font if noone was supplied
	set font ""
	foreach {option val} $args {
	    switch -- $option {
		-font {set font $val}
		default {}
	    }
	}
	if {$font == ""} {
	    lappend args -font default
	}
    }
    
    set object_id [$data(workspace) find withtag $canvas_uid]
    set tags [list $canvas_uid $canvas_id $canvas_rule $canvas_symb $canvas_depth]
    if {"$object_id" == ""} {
	set object_id [eval {$data(workspace) create $type} $coords\
	    {-tags $tags} \
	    $canvargs]
    } else {
	eval {$data(workspace) coords $object_id} $coords
	eval {$data(workspace) itemconfigure $object_id -tags $tags} $canvargs
    }

    if {[lsearch -exact $data(depths) $depth] < 0} {
	lappend data(depths) $depth
    }
    lappend data(currentids) $object_id
    return $object_id
}

proc ::visual::setWindowTitle {title} {
    upvar #0 $::visual::currVarname data
    set win $data(windowid)
    variable visInfo
    set visInfo($win-title) $title
    updateTitle $win
    #::MDIFrame::SetTitle [::MDIFrame::GetClient $$data(windowid)] $title
}


proc ::visual::set_editable {obid attrname} {
    upvar #0 $::visual::currVarname data

    set type [$data(workspace) type $obid]
    if {$type != "text"} {
	::log::log "set_editable called for nontext"
    }

    set tags [$data(workspace) gettags $obid]
    lappend tags "edittext" "editattr,$attrname"

    $data(workspace) itemconfigure $obid -tags $tags
}

proc ::visual::set_resizer {obid corner attrname} {
    upvar #0 $::visual::currVarname data

    set tags [$data(workspace) gettags $obid]
    lappend tags "noselect" "resizer" 
    tags_setproperty tags "resizerCursor" $corner
    tags_setproperty tags "resizerProc" "stdresize"
    tags_setproperty tags "resizerProcParam" [list $corner $attrname]
    $data(workspace) itemconfigure $obid -tags $tags -fill {} -outline {}
}


proc ::visual::createHandle {obid procname attrlist} {
    upvar #0 $::visual::currVarname data

    set tags [$data(workspace) gettags $obid]
    lappend tags "noselect" "resizer"
    tags_setproperty tags "resizerCursor" 8
    tags_setproperty tags "resizerProc" $procname
    tags_setproperty tags "resizerProcParam" $attrlist
    $data(workspace) itemconfigure $obid -tags $tags -fill {} -outline {}
}


proc ::visual::create_fix_inspoint {id kind coords where partname action posattr} {
    upvar #0 $::visual::currVarname data
    variable inspointkind

# id is a node id of a node. A new node could possibly inserted as a brother
# or son of id or could replace id. 
# kind is one of
#    rectangle             4 coordinates supplied.
#    line                  4 coordinates supplied.
#    dot                   2 coordinates supplied.
# More types can be added by calls to ::visual::create_inspoint_kind
# which is explained in _create_inspoint_instargs

#action can be: 
#  i    Insertion-Point only for insertion of new nodes
#  m    Insertion-Point only for moving nodes, both in structure and position
#  b    Insertion-Point for both insertion and moving, default action.
#  p    Insertion-Point only for repositioning, without changing structure.

    if {[catch {set canvtype $inspointkind(canvastype,$kind)}]} {
	error "Unknown inspoint kind: $kind"
    }
    set time [time {
    set canvargs $inspointkind(defaultattrs,$kind)
    set numcoords $inspointkind(numcoords,$kind)
    set onselect $inspointkind(onselect,$kind)
    set setcoords $inspointkind(setcoords,$kind)

    if {[llength $coords] != $numcoords} {
	error "$kind needs [expr $numcoords / 2] points"
    }
    
    if {$setcoords != ""} {
	eval $setcoords
    }
    
    set canvas_id "inspoint#$id"
    set canvas_uid "inspoint,$id,[check_partname $partname],$where"
    set tags [list $canvas_id $canvas_uid "inskind,$kind" "inspoint"]

    if {$posattr != ""} {
	# cs:ISP
        tags_setproperty tags "inssetpos" $posattr
        #set joined_posattr [join $posattr ","]
	#lappend tags "inssetpos,$joined_posattr"
    }

    eval lappend tags [_create_inspoint_instags $::visual::currVarname $id $where 0 0 $action] 

    set object_id [$data(workspace) find withtag $canvas_uid]
    if {"$object_id" == ""} {
	set object_id [eval {$data(workspace) create $canvtype} $coords \
		{-tags $tags} $canvargs] 
    } else {
	eval {$data(workspace) coords $object_id} $coords
	eval {$data(workspace) itemconfigure $object_id -tags $tags} $canvargs
    }

    lappend data(currentids) $object_id
    }]
    set ::canvastime [expr $::canvastime + [lindex $time 0]]
    return $object_id
}

proc ::visual::create_movable_inspoint {id kind coords bbox where partname depth action posattr} {
    upvar #0 $::visual::currVarname data
    variable inspointkind

# id is a node id of a node. A new node could possibly inserted as a brother
# or son of id or could replace id. 
# kind is one of
#    rectangle             4 coordinates supplied.
#    line                  4 coordinates supplied.
#    dot                   2 coordinates supplied.
# More types can be added by calls to ::visual::create_inspoint_kind
# which is explained in _create_inspoint_instargs

#action can be: 
#  i    Insertion-Point only for insertion of new nodes
#  m    Insertion-Point only for moving nodes, both in structure and position
#  b    Insertion-Point for both insertion and moving, default action.
#  p    Insertion-Point only for repositioning, without changing structure.

    if {[catch {set canvtype $inspointkind(canvastype,$kind)}]} {
	error "Unknown inspoint kind: $kind"
    }
    set time [time {
    set canvargs $inspointkind(defaultattrs,$kind)
    set numcoords $inspointkind(numcoords,$kind)
    set onselect $inspointkind(onselect,$kind)
    set setcoords $inspointkind(setcoords,$kind)

    if {[llength $coords] != $numcoords} {
	error "$kind needs [expr $numcoords / 2] points"
    }
    
    if {$setcoords != ""} {
	eval $setcoords
    }
    
    set canvas_id "inspoint#$id"
    set canvas_uid "inspoint,$id,[check_partname $partname],$where"
    set tags [list $canvas_id $canvas_uid "inskind,$kind"]
    if {$bbox != ""} {
	if {[llength $bbox] != 4} {
	    error "Need 2 Points for bbox, if given"
	}
        tags_setproperty tags "bbox" $bbox
	#lappend tags [join [eval list "bbox" $bbox] ","]
    }
    if {$posattr != ""} {
	#cs:ISP
        tags_setproperty tags "inssetpos" $posattr
        lappend tags "inssetpos,$posattr"
    }

    eval lappend tags [_create_inspoint_instags $::visual::currVarname $id $where 1 $depth $action]

    set object_id [$data(workspace) find withtag $canvas_uid]
    if {"$object_id" == ""} {
	set object_id [eval {$data(workspace) create $canvtype} $coords \
		{-tags $tags} $canvargs] 
    } else {
	eval {$data(workspace) coords $object_id} $coords
	eval {$data(workspace) itemconfigure $object_id -tags $tags} $canvargs
    }

    lappend data(currentids) $object_id
    }]
    set ::canvastime [expr $::canvastime + [lindex $time 0]]
    return $object_id
}



proc ::visual::create_reposition {id kind coords bbox partname depth symbolnames posattr origin} {
    upvar #0 $::visual::currVarname data
    variable inspointkind

# id is a node id of a node. A new node could possibly inserted as a brother
# or son of id or could replace id. 
# kind is one of
#    rectangle             4 coordinates supplied.
#    line                  4 coordinates supplied.
#    dot                   2 coordinates supplied.
# More types can be added by calls to ::visual::create_inspoint_kind
# which is explained in _create_inspoint_instargs

    if {[catch {set canvtype $inspointkind(canvastype,$kind)}]} {
	error "Unknown inspoint kind: $kind"
    }
    set time [time {
    set canvargs $inspointkind(defaultattrs,$kind)
    set numcoords $inspointkind(numcoords,$kind)
    set onselect $inspointkind(onselect,$kind)
    set setcoords $inspointkind(setcoords,$kind)

    if {[llength $coords] != $numcoords} {
	error "$kind needs [expr $numcoords / 2] points"
    }
    
    if {$setcoords != ""} {
	eval $setcoords
    }
    
    set canvas_id "inspoint#$id"
    set canvas_uid "inspoint,$id,[check_partname $partname]"
    set tags [list $canvas_id $canvas_uid "inskind,$kind"]
    if {$bbox != ""} {
	if {[llength $bbox] != 4} {
	    error "Need 2 Points for bbox, if given"
	}
        tags_setproperty tags bbox $bbox
	#lappend tags [join [eval list "bbox" $bbox] ","]
    }

    # Attribut zur Speicherung der Position: Hier
    # kann auch origin untergebracht werden!
    if {$posattr != ""} {
        #cs:ISP
        set posattr [concat $posattr $origin]
	#set joined_posattr [join $posattr ","]
	#lappend tags "inssetpos,$joined_posattr"
        tags_setproperty tags "inssetpos" $posattr
    }

    #if {$posattr != ""} {
    #	lappend tags "inssetpos,$posattr"
    #}

    if {$depth > 0} {
	lappend tags "inspointdepth_$depth"
    }

    set insertable ""
    foreach sym $symbolnames {
	if {[info exists ::vl::symb($sym)]} {
	    eval lappend insertable [::vl::possible_rules $::vl::symb($sym)]
	} else {
	    ::log::log "VLRepositionCreate called with nonexisting symbol-name $sym"
	}
    }
	
    set prefix "inspointmp"
    foreach rule $insertable {
	lappend tags "$prefix,$rule"
    }

    set object_id [$data(workspace) find withtag $canvas_uid]
    if {"$object_id" == ""} {
	set object_id [eval {$data(workspace) create $canvtype} $coords \
		{-tags $tags} $canvargs] 
    } else {
	eval {$data(workspace) coords $object_id} $coords
	eval {$data(workspace) itemconfigure $object_id -tags $tags} $canvargs
    }

    lappend data(currentids) $object_id
    }]
    set ::canvastime [expr $::canvastime + [lindex $time 0]]
    return $object_id
}

proc ::visual::_create_inspoint_instags {varname id where movable depth action} {
# where means the following:
#   1   insertion point should insert as a brother of id above id.
#   2   insertion point should insert as a brother of id below id.
#   3   insertion point should insert as a son of id before first element.
#   4   insertion point should replace id.
#   5   insertion point should insert as a son of id as new last element.

#action can be: 
#  i    Insertion-Point only for insertion of new nodes
#  m    Insertion-Point only for moving nodes, both in structure and position
#  b    Insertion-Point for both insertion and moving, default action.
#  p    Insertion-Point only for repositioning, without changing structure.

# Returns tags that specify insertion point characteristics
# relevant for selection and application of an insertion point:
# For each possible production, the insertion point has a inspoint,<rulenumber> tag.
# Additionally, it has a "insargs"-tag containing three of the four parameters
# required for the insert-command, separated by commas.

    if {$movable} {
	set prefix "inspointm$action"
    } else {
	set prefix "inspointf$action"
    }

    set tags ""
    if {$depth > 0} {
	lappend tags "inspointdepth_$depth"
    }

    if {$where < 3} {
	set father [::vl::getFather $id]
	if {$father == ""} {
	    return $tags           ;# No Father means root. No brother of root allowed.
	}
	set fathercode [::vl::getCode $father]
	set fathertype [::vl::ruletype $fathercode]
	if {$fathertype != 1} {
	    return $tags   ;# For now, only in lists can be inserted.
	}
	set insertable [::vl::possible_rules [::vl::listelem [::vl::lhs $fathercode]]]

	# Construct tags
	foreach rule $insertable {
	    lappend tags "$prefix,$rule"
	}
	set index [lsearch -exact [::vl::getDesc $father] $id]
	if {$where == 2} { incr index }
	lappend tags [join [eval list "insargs" $father $index 0] ","]
	return $tags
    } elseif {$where == 4} {
	set father [::vl::getFather $id]
	if {$father == ""} {
	    return $tags           ;# No Father means root. Root Production is unique and cannot be replaced.
	}
	set code [::vl::getCode $id]
	set insertable [::vl::possible_rules [::vl::lhs $code]]

	# Construct tags
	foreach rule $insertable {
	    lappend tags "$prefix,$rule"
	}
	set index [lsearch -exact [::vl::getDesc $father] $id]
	lappend tags [join [eval list "insargs" $father $index 1] ","]
	return $tags
    } else {
	set code [::vl::getCode $id]
	set type [::vl::ruletype $code]
	if {$type != 1} {
	    return $tags   ;# For now, only in lists can be inserted.
	}
	set insertable [::vl::possible_rules [::vl::listelem [::vl::lhs $code]]]

	foreach rule $insertable {
	    lappend tags "$prefix,$rule"
	}
	set index 0
	if {$where == 5} {
	    set index [llength [::vl::getDesc $id]]
	}

	lappend tags [join [eval list "insargs" $id $index 0] ","]
	return $tags
    }
}

proc ::visual::set_inspoint_proc {inspointid insertcmd insertargs} {
    upvar #0 $::visual::currVarname data
    
    set tags [$data(workspace) gettags $inspointid]
    lappend tags "insertcmd,$insertcmd"
    $data(workspace) itemconfigure $inspointid -tags $tags

    set data(vargs,$inspointid) $insertargs
}

proc ::visual::setPosAttrProc {inspointid posattrproc posattrprocargs} {
    upvar #0 $::visual::currVarname data
    set tags [$data(workspace) gettags $inspointid]
    tags_setproperty tags "posAttrProc" [concat $posattrproc $posattrprocargs]
    $data(workspace) itemconfigure $inspointid -tags $tags
}

proc ::visual::constrain_inspoint {inspointid root tree} {
    upvar #0 $::visual::currVarname data
    
    set tags [$data(workspace) gettags $inspointid]
    lappend tags "inscontext,$tree,$root"
    $data(workspace) itemconfigure $inspointid -tags $tags
}
    

proc ::visual::create_drag {id partname type kind attrname coords associatecmd associateargs} {
    upvar #0 $::visual::currVarname data

    if {($type == "rectangle" || $type == "oval") && [llength $coords] == 2} {
	set halo 2
	set x [lindex $coords 0]
	set y [lindex $coords 1]
	
	set coords [list [expr $x - $halo] [expr $y - $halo] [expr $x + $halo] [expr $y + $halo]]
    }

    set canvas_id "drag#$id"
    set canvas_uid "drag,$id,[check_partname $partname]"
    set tags [list $canvas_id $canvas_uid "dragkind,$kind" "dragattr,$attrname" "drag"]
    if {$associatecmd != ""} {
	lappend tags "dragcmd,$associatecmd"
    }
    set object_id [$data(workspace) find withtag $canvas_uid]
    if {"$object_id" == ""} {
	set object_id [eval {$data(workspace) create $type} $coords \
		{-tags $tags -width 1 -outline ""}]
    } else {
	eval {$data(workspace) coords $object_id} $coords
	eval {$data(workspace) itemconfigure $object_id -tags $tags}
    }

    lappend data(currentids) $object_id

    set data(vargs,$object_id) $associateargs
    return $object_id
}

proc ::visual::droptarget {property obid kind dropargs} {
    upvar #0 $::visual::currVarname data

    set tags [$data(workspace) gettags $obid]
    lappend tags "drop,$kind" "prop,$property" "noselect"
    $data(workspace) itemconfigure $obid -tags $tags

    set data(vargs,$obid) $dropargs
}

proc ::visual::do_contextedit {id label attribute value} {
    upvar #0 $::visual::currVarname data

    lappend data(contextmenu,attr,$id) $label $attribute $value
}

proc ::visual::do_contexteditcmd {id label cmd} {
    upvar #0 $::visual::currVarname data

    lappend data(contextmenu,cmd,$id) $label $cmd
}


proc ::visual::addEditContext {visualContext editContext} {
    upvar #0 $::visual::currVarname data

    lappend data(additionalEditContext,$visualContext) $editContext
}
